 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef BIGARRAY_H
#define BIGARRAY_H


#include <vector>
#include <cassert>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/mman.h>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include <sstream>


using namespace std;

template<class DataType>
class BigArrayIterator; 

template<class DataType>
class OverallIterator; 

/***
 * This class represent a very large array stored using a memory mapped file.
 * In theory, it should support very, very large files, up to 2**32 * sizeof(DataType)
 * which could be several times 4 GigaBytes.
 *
 * In practice, this would most certainly fail, because the code would be buggy,
 * but also because the file system may not be able to handle it. 
 *
 * author: Daniel Lemire, March 2004
 * 
 */
template<class DataType>
class BigArray : public MultidimensionalArray<DataType> {
        public:
               BigArray(vector<int> shape, string filename = "BigArrayTemp.bin") throw(std::ifstream::failure);
                BigArray(int shape[], int NumberOfDimensions, string filename = "BigArrayTemp.bin") throw(std::ifstream::failure);
                
                BigArray(const BigArray& bt): MultidimensionalArray<DataType>(bt), mFileDescriptor(bt.mFileDescriptor), mData(bt.mData), mFileName(bt.mFileName) {}

                 /**
                * return the name of the cube, typically a filename where the data
                * is stored, but could also be something like mysql:database/432:table2
                */
                string getName() const {return mFileName;}

                virtual ~BigArray();

                //
                inline DataType get(const int  pos[]) const;
                inline DataType get(const vector<int>&  pos) const;

                // add delta and return new value
                inline DataType increment(const vector<int>&  pos, DataType delta) const;
                inline DataType increment(const int  pos[], DataType delta) const;
 
                inline void put(const vector<int>& pos, DataType value);
                inline void put(const int  pos[], DataType value);

		// undefinedAt should be based on a auxiliary bitmap
		inline bool undefinedAt(const int pos[]) const { return false;} // these are broken -ofk
		inline bool undefinedAt( const vector<int> & pos) const { return false;}

                // reset the content of the file to all zeroes
                // (not need to call this generally)
                virtual void fillWithZeroes();
                
                void sync() {
                        msync(mData, this->getVolume(), MS_SYNC);
                }

                
                // if you manually closed the mapping, you can reopen it
                virtual void open() throw(std::ifstream::failure);
                
                // if you want, you can manually close the mapping
                virtual void close();

                 /*BigArrayIterator<DataType>*/ typename MultidimensionalArray<DataType>::IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
                        return typename MultidimensionalArray<DataType>::IteratorPtr(new BigArrayIterator<DataType>(*this,start,end));
                 }
                 
                 
                 typename MultidimensionalArray<DataType>::IteratorPtr iterator() const {
                           return typename MultidimensionalArray<DataType>::IteratorPtr(new OverallIterator<DataType>(*this));
                 }
    

 
         private:
                virtual MultidimensionalArray<DataType>* createNew(vector<int> shape, string file)  const  throw(std::ifstream::failure);// {return new BigArray<DataType>(shape,file);}
                // actually go to the file stream and write zeroes everywhere
                void blankFile();
                
                int mFileDescriptor; //file descriptor
                DataType *mData;// pointer to data

                string mFileName;
                
                inline DataType fastget(const int64  pos) const {
                        return mData[pos];
                }
                
                friend class OverallIterator<DataType>;

                

};


template <class DataType>
BigArray<DataType>::BigArray(vector<int> Shape, string filename) throw(std::ifstream::failure)
  : MultidimensionalArray<DataType>(Shape), mFileDescriptor(-1), mData(NULL), mFileName(filename){
  open();
}
template <class DataType>
BigArray<DataType>::BigArray(int Shape[], int NumberOfDimensions, string filename) throw(std::ifstream::failure)
  : MultidimensionalArray<DataType>(Shape, NumberOfDimensions), mFileDescriptor(-1), mData(NULL), mFileName(filename){
  open();
}

template <class DataType>
MultidimensionalArray<DataType>* BigArray<DataType>::createNew(vector<int> shape, string file)  const  throw(std::ifstream::failure) {return new BigArray<DataType>(shape,file);}

template <class DataType>
BigArray<DataType>::~BigArray() {
  if(this->getCount() == 1) close();
}

template <class DataType>
void BigArray<DataType>::open() throw(std::ifstream::failure) {
  close();
  FileUtil::mkdir_parentsonly(mFileName.c_str());
  //cout << " getVolume() = " << getVolume() << endl;
  //if the file doesn't exist, or if it exists and the file size is zero....make it the right size
  if  (!FileUtil::fileExists(mFileName.c_str()) || 
      (FileUtil::getFileSize(mFileName.c_str()) < this->getVolume()*sizeof(DataType)))   {
    // make sure the path the file exists
    //                
    const char * path = mFileName.c_str();
    char * shorter_path = strdup(path);
    char * p = strrchr(shorter_path, '/');// find the last occurance of '/'
    if(p == 0 || p == path) free(shorter_path);
    else {
      p[0] = 0;// set blank
      if(verbose) cout << "Creating directory "<<shorter_path << endl;
      int returnvalue = FileUtil::mkdir_parents(shorter_path);// should check return value
      if(verbose) cout << " Return value = " << returnvalue << endl;
      free(shorter_path);
    }
    // create the file
    blankFile();
  }
  FileUtil::mkdir_parentsonly(mFileName.c_str());
  mFileDescriptor = ::open(mFileName.c_str(), O_RDWR | O_CREAT);
  if(mFileDescriptor == -1 ) {
      stringstream strs;
      strs << "Failed to open file "<< mFileName<< " in R/W mode error code is " << errno<< " (see errno.h)";
      //string errormsg;
      //strs >>  errormsg ; 
      throw ifstream::failure(strs.str());
  }
  // I strongly suspect that if the volume is too big, the next line will fail
  // For now, just make the code fail until someone can check this out
  if((uint64) (uint) (sizeof(DataType)* (this->getVolume())) != (sizeof(DataType)*(this->getVolume())))
    throw length_error("bigarray_open");
  mData = (DataType *)mmap((caddr_t)0, sizeof(DataType)*(this->getVolume()), 
      PROT_READ | PROT_WRITE, MAP_SHARED, mFileDescriptor, 0);
  if ( mData == MAP_FAILED) {
      stringstream strs;
      strs << "Failed to map file "<< mFileName<< " error code is " << errno<< " (see errno.h)";
      //string errormsg;
      //strs >>  errormsg ; 
      throw ifstream::failure(strs.str());
  }
}



template <class DataType>
void BigArray<DataType>::blankFile() {
  //cout << " blanking the file " << endl;
  DataType * Big = new DataType[this->mShape[0]];
  memset(Big,0,this->mShape[0] * sizeof(DataType));
  int NumberOfTimes = 1;
  for(uint i = 1 ; i < this->mShape.size() ; ++i ) NumberOfTimes *= this->mShape[i];
  if(verbose) cout << "Blanking "<<mFileName << endl;
  ofstream FileStream( mFileName.c_str(),std::ios::binary |  std::ios::trunc);
  if(!FileStream.is_open()) {
    cout << " I could not create/write to " << mFileName << endl;
  }
  assert(FileStream.is_open());
  assert(!FileStream.fail());
  //FileStream.seekp(0);
  for(int times = 0; times < NumberOfTimes; ++times) 
    FileStream.write((char *) Big,this->mShape[0] * sizeof(DataType));
  assert(! FileStream.fail());
  FileStream.close();
  delete[] Big;
}


template <class DataType>
void BigArray<DataType>::close() {
  if (mData != NULL)  {
    munmap(mData, sizeof(DataType)*(this->getVolume()));
    mData = NULL;
    ::close(mFileDescriptor);
  }
}



template <class DataType>
DataType BigArray<DataType>::get(const vector<int>& pos) const {
  uint64 offset = this->computeOffset(pos);
  //cout << " offset = "<< offset << endl;
  const DataType answer = mData[offset];
  //cout << " done " <<endl;
  return answer;
}

template <class DataType>
DataType BigArray<DataType>::get(const int pos[]) const {
  uint64 offset = this->computeOffset(pos);
  //cout << " offset = "<< offset << endl;
  const DataType answer = mData[offset];
  //cout << " done " <<endl;
  return answer;
}

template <class DataType>
DataType BigArray<DataType>::increment(const vector<int>& pos, DataType delta) const {
  uint64 offset = this->computeOffset(pos);
  const DataType answer = (mData[offset]+=delta);
  return answer;
}

template <class DataType>
DataType BigArray<DataType>::increment(const int pos[], DataType delta) const {
  uint64 offset = this->computeOffset(pos);
  const DataType answer = (mData[offset]+=delta);
  return answer;
}

template <class DataType>
void BigArray<DataType>::put(const vector<int>& pos, DataType value) {
  uint64 offset = this->computeOffset(pos);
  *(mData + offset) = value;
}

template <class DataType>
void BigArray<DataType>::put(const int pos[], DataType value) {
  uint64 offset = this->computeOffset(pos);
  *(mData + offset) = value;
}



template <class DataType>
void BigArray<DataType>::fillWithZeroes()
{
  //fill memory with zeroes and sync file and memory.
  memset(mData,0,this->getVolume()*sizeof(DataType));
  //for (uint64 index = 0; index < this->getVolume(); index++)
  //  *(mData + index) = 0;
  msync(mData, this->getVolume(), MS_SYNC);
}


/**
* This is also used by internalbigarray
*/
template <class DataType>
class BigArrayIterator:public MultidimensionalIterator<DataType> {

  public: 

    BigArrayIterator(const MultidimensionalArray<DataType> & owner,const vector<int>& start, const vector<int>& end) : MultidimensionalIterator<DataType>(),
      mStart(start), mEnd(end),mIndex(start),
      mCurrentValue(0),mOwner(owner), mSize(start.size()){
        mIndex[0] -= 1;// start right before the area
    }

    BigArrayIterator(const BigArrayIterator<DataType> & other):MultidimensionalIterator<DataType>(),
      mStart(other.mStart), mEnd(other.mEnd), mIndex(other.mIndex), 
      mCurrentValue(other.mCurrentValue), mOwner(other.mOwner), mSize(other.mSize)
    {}
    /**
     * This is where the code needs to get smarter.
     */
/*    inline bool slownext() {
      if(! increment ()) return false;
      while((mCurrentValue = mOwner.get(mIndex)) == 0) 
        if(! increment ()) return false;
      return true;
    }*/

    /**
    * this is manually inlined
    */
    inline bool next() {
      for(uint dim = 0; dim < mSize; ++dim) {
        if(++mIndex[dim] >= mEnd[dim]) {
          mIndex[dim] = mStart[dim];
        } else {
          while((mCurrentValue = mOwner.get(mIndex)) == 0) {
              bool ok = false;
              for(uint dim = 0; dim < mSize; ++dim) {
                if(++mIndex[dim] >= mEnd[dim]) {
                  mIndex[dim] = mStart[dim];
                } else {
                  ok = true;
                  break;
                }
              }
              if(! ok) return false;                   
          }
          return true;
        }
      }
      return false;
    }
    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key()  {
      return mIndex;
    }
    BigArrayIterator * clone() const {return new BigArrayIterator(*this);}

  protected:
    inline bool increment() {
      for(uint dim = 0; dim < mSize; ++dim) {
        if(++mIndex[dim] >= mEnd[dim]) {
          mIndex[dim] = mStart[dim];
        } else {
          return true;
        }
      }
      return false;
    }

      
    vector<int> mStart, mEnd, mIndex;
    DataType mCurrentValue;
    const MultidimensionalArray<DataType> & mOwner;
    const uint mSize;

};



template <class DataType>
class OverallIterator:public MultidimensionalIterator<DataType> {

  public: 

    OverallIterator(const BigArray<DataType> & owner) : MultidimensionalIterator<DataType>(), mOwner(owner), mVolume(mOwner.getVolume()),mIndex(-1), mCurrentValue(0), mCurrentPos(owner.getShape().size()){

    }

    OverallIterator(const OverallIterator<DataType> & other):MultidimensionalIterator<DataType>(),mOwner(other.mOwner), mVolume(other.mVolume),mIndex(other.mIndex), mCurrentValue(other.mCurrentValue),mCurrentPos(other.mCurrentPos)
    {}
    
    inline bool next() {
      if(++mIndex >= mVolume) return false;
      while((mCurrentValue = mOwner.fastget(mIndex)) == 0) 
        if(++mIndex >= mVolume) return false;
      return true;
    }

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key() {
      mOwner.reverseoffset(mIndex,mCurrentPos);
      return mCurrentPos;      
    }
    
    OverallIterator<DataType> * clone() const {return new OverallIterator<DataType>(*this);}
    
  protected:
    
    const BigArray<DataType> & mOwner;
    const uint64 mVolume;
    uint64 mIndex;
    DataType mCurrentValue;
    vector<int> mCurrentPos;


};


#endif
