 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
 
#ifndef BTreeCube_H
#define BTreeCube_H

#include <string>
#include <fstream>
#include <cassert>
#include <vector>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <map>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include <villa.h> // QDBM, see http://qdbm.sourceforge.net/, please download and install!, RPMs are available for RedHat-like systems
// Daniel: QDBM has a C++ API, but I found that it was somewhat limited (I complained
// to the author and maybe it is fixed now...)
// Also, we use villa because it has sorted keys, this means we can

using namespace std;


/**
* Note that we use a little endian approach here!!! When you increment,
* you have to increment the right most value first.
*/
static inline int vlintarraycompare(const char *aptr, int asiz, const char *bptr, int bsiz){
  //assert(aptr && asiz >= 0 && bptr &&  asiz == bsiz);
  const int * aintptr = (const int *) aptr;
  const int * bintptr = (const int *) bptr;  
  for(uint i = 0; i < asiz/sizeof(int); i++) {
    if(aintptr[i] != bintptr[i]) 
            return aintptr[i] - bintptr[i];
  }
  return 0;
}





VLCFUNC VL_CMPINTARRAY = vlintarraycompare;

template<class DataType>
class BTreeCubeIterator; 
template<class DataType>
class BTreeCubeOverallIterator;
template<class DataType>
class BTreeCubeSliceIterator;
template<class DataType>
class BTreeCubeFirstDimensionSliceIterator;




/***
 * Consider a "sparse cube", that is a map from coordinate tuples to values
 * It is easy to store such a "sparse cube" in a B-tree using a DBM. 
 * This code maps tuples to "C-like" indexes and use these indexes
 * as keys in the B-Tree. By derivation, you can support other indexes.
 *
 * See also ubtreecube.h (to come!).
 *
 * TODO: write an iterator over a range [done?].
 *
 * author: Daniel Lemire, February 2005
 * 
 */
template<class DataType>
class BTreeCube : public MultidimensionalArray<DataType> {

 
  public:

    /**
     * Create a cube with chunks of a given shape. The data will be stored in 
     * a file (you must have write permissions to it, it will be created
     * if it doesn't exist). 
     */
    BTreeCube( int NumberOfDimensions, string filename = "BTreeCubeTemp.bin") throw(std::ifstream::failure);/*int shape[],*/
    // the actual content of shape is irrelevant, only the number of dimensions matter!
    BTreeCube(vector<int> shape, string filename ="BTreeCubeTemp.bin") throw(std::ifstream::failure);


    BTreeCube(const BTreeCube& bt): MultidimensionalArray<DataType>(bt), mFileName(bt.mFileName), mVilla(bt.mVilla) {
      //cout << "copy constructor for BTreeCube" << endl;
    }
    
    virtual ~BTreeCube();
    
    /**
    * return the name of the cube, typically a filename where the data
    * is stored, but could also be something like mysql:database/432:table2
    */
    string getName() const {return mFileName;}

    virtual int tempcode() const { return 2;}
   
    /**
     * In very rare cases, you might want to close the cube to 
     * release system resources. (Destructor will do it for you...)
     */
    void close();
   
    /**
     * If you ever happened to close the cube, you might need to open it.
     */
    void open() throw(std::ifstream::failure);
    /**
    * erase all 
    */
    void fillWithZeroes();
    
    /**
     * Query the cube. If the cell is not allocated, returns 0.
     */
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;

    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;

    /**
     * put a value at a given cell. If the value is 0, then 
     * it should delete the cell.  (Owen is uncomfortable with this deletion.)
     * (Daniel doesn't feel strongly about it, I merely documented the behavior.
     * if we change it, then we need a different call for deleting entries.)
     */
    inline void put(const vector<int>& pos, DataType value);
    inline void put(const int pos[], DataType value);

    void dump(void);
    void sync() {
            if (! vlsync(mVilla)) {
           cerr  << __FILE__ << ":"<< __LINE__ << " I didn't manage to sync the data with "<< mFileName <<": " <<dperrmsg(dpecode)<< endl;  }
    }

    /**
    * Now, I think we support ranges correctly.
    */
    typename MultidimensionalArray<DataType>::
      IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new BTreeCubeIterator<DataType>(*this,start,end));
    }

    typename MultidimensionalArray<DataType>::IteratorPtr iterator(void) const {
 	return typename MultidimensionalArray<DataType>::
	  IteratorPtr(new BTreeCubeOverallIterator<DataType>(*this));
     }
     

    typename MultidimensionalArray<DataType>::IteratorPtr sliceIterator( int dim, int slice_at) const {
        if(dim == 0)
                return new BTreeCubeFirstDimensionSliceIterator<DataType>(*this,slice_at);
 	return new BTreeCubeSliceIterator<DataType>(*this,dim,slice_at);
     }

     
     int getNumberOfRecords() {return vlrnum(mVilla);}

  protected:     
  

  private:
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape, string file) const  throw(std::ifstream::failure); //{return new BTreeCube<DataType>(shape,file);}
    
    enum{verbose=false,syncaftereachput=false,checkfornulls=false};    
    friend class BTreeCubeIterator<DataType>; 
    friend class BTreeCubeOverallIterator<DataType>; 
    friend class BTreeCubeSliceIterator<DataType>;
    friend class BTreeCubeFirstDimensionSliceIterator<DataType>; 
    string mFileName;
    VILLA *mVilla;    

};

template<class DataType>
BTreeCube<DataType>::BTreeCube(vector<int> shape, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(shape),    mFileName(filename), mVilla(NULL)  {
  // we ensure the shape is really zeroed
  for (uint k = 0; k < this->mShape.size(); ++k) this->mShape[k] = this->INVALID_SIZE;
  open();

}
template<class DataType>
BTreeCube<DataType>::BTreeCube(int NumberOfDimensions, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)),    mFileName(filename), mVilla(NULL)  {
  open();

}

template<class DataType>
MultidimensionalArray<DataType>* BTreeCube<DataType>::createNew(vector<int> shape, string file) const  throw(std::ifstream::failure){return new BTreeCube<DataType>(shape,file);}

template<class DataType>
void BTreeCube<DataType>::dump(void) {
  typename MultidimensionalArray<DataType>::IteratorPtr iter = iterator(); 
  while (iter->next()) {
    cout << "value " << iter->value() << " at ";
    for (uint i=0; i < this->mShape.size(); ++i)
      cout << " " << iter->key()[i];
    cout << endl;
    put( iter->key(), iter->value());
  }
  delete iter;
}



 
template<class DataType>
void BTreeCube<DataType>::close() {
        if(mVilla) {
           if(!vlclose(mVilla)){
             cerr << __FILE__ << ":"<< __LINE__ <<" I failed to close " << mFileName << endl;       
             cerr << dperrmsg(dpecode)<<endl;
             // here, the user cannot do much, so I'm not throwing an exception. Hope for the best!
           }
        }
        mVilla = NULL;
}


template<class DataType>
BTreeCube<DataType>::~BTreeCube() {
    if(this->getCount() == 1) close();
} 
 
template<class DataType>
void BTreeCube<DataType>::open() throw(ifstream::failure) {
  close();
  FileUtil::mkdir_parentsonly(mFileName.c_str());
  if( ! (mVilla = vlopen(mFileName.c_str(), VL_OWRITER | VL_OCREAT, VL_CMPINTARRAY))) {
    cerr  << __FILE__ << ":"<< __LINE__ << " [BTreeCube::open] I failed to open " << mFileName << endl;
    cerr << " [BTreeCube::open] The file might be corrupted: not a QDBM format. "<< endl;
    cerr << " [BTreeCube::open] You may have to delete it. "<< endl;
    stringstream strs;
    strs << " [BTreeCube::open] Failed to open file "<< mFileName<< ", error might be QDBM related. I'm throwing an ifstream::failure!";
    strs << dperrmsg(dpecode);
    //string errormsg;
    //strs >>  errormsg ; 
    throw ifstream::failure(strs.str());
  }
  //  vlsettuning(mVilla,14096,0,0,0);// vlsettuning seems to make no difference
}

template<class DataType>
void BTreeCube<DataType>::fillWithZeroes() {
        close();
        if( ! (mVilla = vlopen(mFileName.c_str(), VL_OWRITER | VL_OTRUNC, VL_CMPINTARRAY))) {
           cerr  << __FILE__ << ":"<< __LINE__ << "[BTreeCube::fillWithZeroes] I failed to open " << mFileName << endl;
        }

}

                

template<class DataType>
DataType BTreeCube<DataType>::get(const vector<int>& pos) const {
        char * val = NULL;
        //const uint64 key = index(pos);
        //if(verbose > 5) cout << " querying key : " <<key << endl;
        /* Note that & pos[0] is safe as long as pos is not empty...*/
	/* Owen: this is very scary to me.  I do not trust that vector<int> has its
           data storage contiguous, like an array. That would be a good implementation
           ...but  Are you SURE this is always okay across all STL implementations? */
        /* Daniel: no, this is absolutely safe. I am SURE, SURE, SURE!!!*/
	// Owen's paranoia...is the last element where we think it is?....
	//assert(  ( ( int  *) & pos[0])+(pos.size()-1) == (int *) & pos[ pos.size()-1]);

        if(!(val = vlget(mVilla, (const char *) & pos[0], sizeof(int)*pos.size(), NULL))) {
                return 0;// no value was found
        }
        DataType answer(*(DataType*)val);
        free(val);
        return answer;
}

template<class DataType>
DataType BTreeCube<DataType>::get(const int pos[]) const {
        char * val = NULL;
        //const uint64 key = index(pos);
        //    if(verbose > 5) cout << " querying key : " <<key << endl;
        //
        if(!(val = vlget(mVilla, (const char *) & pos[0], sizeof(int)*this->mShape.size(), NULL))) {
                return 0;// no value was found
        }
        DataType answer(*(DataType*)val);
        free(val);
        return answer;
}

template<class DataType>
void BTreeCube<DataType>::put(const vector<int>& pos, DataType value) {
#if 0
  cout << "put computes index using pos = " ;
  for (int i=0; i < pos.size(); ++i)
    cout << " " << pos[i];
  cout << endl;
#endif

  //const uint64 key = index(pos);
//  if(verbose > 5) cout << " querying key : " <<key << endl;
  if( value == 0) {
          // no error handling because maybe the key doesn't exist and we are storing zero there!
          vlout(mVilla, (const char *) &pos[0], pos.size()*sizeof(int));
          return;
  }
  if (! vlput(mVilla, (const char *) &pos[0], pos.size()*sizeof(int), (const char *)&value, sizeof(value), VL_DOVER)) {
          cerr  << __FILE__ << ":"<< __LINE__ << " I didn't manage to store the data in "<< mFileName <<": " <<dperrmsg(dpecode)<< endl; 
  }
}



template<class DataType>
void BTreeCube<DataType>::put(const int pos[], DataType value) {
  if( value == 0) {
          // no error handling because maybe the key doesn't exist and we are storing zero there!
          vlout(mVilla, (const char *) &pos[0], sizeof(int)*this->mShape.size());
          return;
  }
  if (! vlput(mVilla, (const char *) &pos[0], sizeof(int)*this->mShape.size(), (const char *)&value, sizeof(value), VL_DOVER)) {
          cerr  << __FILE__ << ":"<< __LINE__ << " I didn't manage to store the data in "<< mFileName <<": " <<dperrmsg(dpecode)<< endl; 
  }
}


template<class DataType>
bool BTreeCube<DataType>::undefinedAt(const int pos[]) const {
        //const uint64 key = index(pos);
        return !vlget(mVilla, (const char *) & pos[0], sizeof(int)*this->mShape.size(), NULL);
}



template<class DataType>
bool BTreeCube<DataType>::undefinedAt(const vector<int>& pos) const {
        //const uint64 key = index(pos);
        return !vlget(mVilla, (const char *) & pos[0], sizeof(int)*pos.size(), NULL);
}


/**
* Object btree is completed
*/


/**************************************
* range iterator
* Optimized to go through a range (defined as the subcube
* spanned by any two cell). The last cell is "not included"
* so that you'd go through an array of length n by specifying
* 0 and n.
*
*************************************/

template <class DataType>
class BTreeCubeIterator:public MultidimensionalIterator<DataType> {

  public: 
    BTreeCubeIterator(const BTreeCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
      MultidimensionalIterator<DataType>(),mStart(start), mEnd(end),mOwner(owner),  mKey(mStart),
      mCurrentValue(0){
        mKey[mKey.size()-1] -= 1;
    }

    BTreeCubeIterator(const BTreeCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      mStart(other.mStart), mEnd(other.mEnd), mOwner(other.mOwner), mKey(other.mKey), 
      mCurrentValue(other.mCurrentValue)
    {}

    
    inline bool next() {
        do {
         for(uint dim = mKey.size()-1;;--dim) {
           ++mKey[dim];
           if(mKey[dim] >= mEnd[dim]) {
             if(dim == 0) return false;
             mKey[dim] = mStart[dim];
             continue;
           }
           break;
         }	 
         if(! vlcurjump(mOwner.mVilla, (const char *) &mKey[0], sizeof(int)*mKey.size(), VL_JFORWARD)) return false;
         int *pkey = (int *) vlcurkey(mOwner.mVilla,NULL);
         DataType *pvalue =  (DataType *) vlcurval(mOwner.mVilla,NULL);
         mCurrentValue = *pvalue;
         /**
         * use memcpy to speed up next line !!!
         */
         //for(uint k = 0; k < mKey.size(); ++k) mKey[k] = pkey[k];
         memcpy(&mKey[0],pkey,mKey.size()*sizeof(int));
         free(pvalue);
         free(pkey);
        } while(! isInRange());
	return true;  
    }    

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key()  {
      return mKey;
    }
    
    BTreeCubeIterator<DataType> * clone() const {return new BTreeCubeIterator<DataType>(*this);}
  protected:

  
    inline bool isInRange() {
            for(uint k = 0 ; k < mStart.size(); ++k) {
                    if((mKey[k] < mStart[k]) || (mEnd[k] > 0 && mKey[k]>= mEnd[k] )) {
                            return false;
                    }
            }
            return true;
    }
    vector<int> mStart, mEnd;
    const BTreeCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
};



/**************************************
* Overall iterator
* Optimized to go through the entire cube
*
*************************************/
template <class DataType>
class BTreeCubeOverallIterator:public MultidimensionalIterator<DataType> {

  public: 

    BTreeCubeOverallIterator(const BTreeCube<DataType> & owner) :
      MultidimensionalIterator<DataType>(),mKey(owner.getShape().size(),0),
      mCurrentValue(0),mOwner(owner){
        mKey[mKey.size()-1] = -1;
    }

    BTreeCubeOverallIterator(const BTreeCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      mKey(other.mKey),  mCurrentValue(other.mCurrentValue), mOwner(other.mOwner)
    {}

    
    inline bool next() {
         for(uint dim = mKey.size()-1;;--dim) {
           ++mKey[dim];
           if(mKey[dim] >= INT_MAX) {
             if(dim == 0) return false;
             mKey[dim] = 0;
             continue;
           }
           break;
         }
         if(! vlcurjump(mOwner.mVilla, (const char *) &mKey[0], sizeof(int)*mKey.size(), VL_JFORWARD)) return false;
         int *pkey = (int *) vlcurkey(mOwner.mVilla,NULL);
         DataType *pvalue = (DataType *) vlcurval(mOwner.mVilla,NULL);
         mCurrentValue = *pvalue;
         /**
         * use memcpy to speed up next line !!!
         */
         //for(uint k = 0; k < mKey.size(); ++k) mKey[k] = pkey[k];
         memcpy(&mKey[0],pkey,mKey.size()*sizeof(int));
         free(pvalue);
         free(pkey);
	 return true;  
    }
    
    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key() {
      return mKey;
    }
    
    BTreeCubeOverallIterator<DataType> * clone() const {return new BTreeCubeOverallIterator<DataType>(*this);}
  protected:


    vector<int> mKey;
    DataType mCurrentValue;
    const BTreeCube<DataType> & mOwner;

};


/**************************************
* Slice iterator
*
*************************************/

template <class DataType>
class BTreeCubeSliceIterator:public MultidimensionalIterator<DataType> {

  public: 

    BTreeCubeSliceIterator(const BTreeCube<DataType> & owner, int dim, int at) :
      MultidimensionalIterator<DataType>(),mKey(owner.getShape().size(),0),
      mCurrentValue(0),mOwner(owner), mDim(dim), mAt(at), mStart(mKey), mEnd(owner.getShape()){
        mKey[mDim] = mAt;
        if((uint)mDim == mKey.size()-1) 
                --mKey[mKey.size()-2];
        else 
                --mKey[mKey.size()-1];
        mStart[mDim] = mAt;
        mEnd[mDim] = mAt+1;
    }

    BTreeCubeSliceIterator(const BTreeCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      mKey(other.mKey),  mCurrentValue(other.mCurrentValue), mOwner(other.mOwner), mDim(other.mDim), mAt(other.mAt),
      mStart(other.mStart), mEnd(other.mEnd)
    {}

    inline bool next() {
        do {
         for(uint dim = mKey.size()-1;;--dim) {
           ++mKey[dim];
           if(mKey[dim] >= mEnd[dim]) {
             if(dim == 0) return false;
             mKey[dim] = mStart[dim];
             continue;
           }
           break;
         }	 
         if(! vlcurjump(mOwner.mVilla, (const char *) &mKey[0], sizeof(int)*mKey.size(), VL_JFORWARD)) return false;
         int *pkey = (int *) vlcurkey(mOwner.mVilla,NULL);
         DataType *pvalue =  (DataType *) vlcurval(mOwner.mVilla,NULL);
         mCurrentValue = *pvalue;
         memcpy(&mKey[0],pkey,mKey.size()*sizeof(int));
         free(pvalue);
         free(pkey);
        } while( mKey[mDim]!=mAt);
	return true;  
    }

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key() {
      return mKey;
    }
    
    BTreeCubeSliceIterator<DataType> * clone() const {return new BTreeCubeSliceIterator<DataType>(*this);}
  protected:

  
    vector<int> mKey;    
    DataType mCurrentValue;
    const BTreeCube<DataType> & mOwner;
    int mDim, mAt;
    vector<int> mStart, mEnd;

};

/**************************************
* Slice iterator (over dimension 0)
*
*************************************/

template <class DataType>
class BTreeCubeFirstDimensionSliceIterator:public MultidimensionalIterator<DataType> {

  public: 

    BTreeCubeFirstDimensionSliceIterator(const BTreeCube<DataType> & owner, int at) :
      MultidimensionalIterator<DataType>(),mKey(owner.getShape().size(),0),
      mCurrentValue(0),mOwner(owner),  mAt(at), mStart(mKey), mEnd(owner.getShape()){
        mKey[0] = mAt;
        mStart[0] = mAt;
        mEnd[0] = mAt+1;
    }

    BTreeCubeFirstDimensionSliceIterator(const BTreeCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      mKey(other.mKey),  mCurrentValue(other.mCurrentValue), mOwner(other.mOwner),  mAt(other.mAt),
      mStart(other.mStart), mEnd(other.mEnd)
    {}

    inline bool next() {
         for(uint dim = mKey.size()-1;;--dim) {
           ++mKey[dim];
           if(mKey[dim] >= mEnd[dim]) {
             if(dim == 0) return false;
             mKey[dim] = mStart[dim];
             continue;
           }
           break;
         }	 
         if(! vlcurjump(mOwner.mVilla, (const char *) &mKey[0], sizeof(int)*mKey.size(), VL_JFORWARD)) return false;
         int *pkey = (int *) vlcurkey(mOwner.mVilla,NULL);
         DataType *pvalue =  (DataType *) vlcurval(mOwner.mVilla,NULL);
         mCurrentValue = *pvalue;
         memcpy(&mKey[0],pkey,mKey.size()*sizeof(int));
         free(pvalue);
         free(pkey);
         if(mKey[0]!=mAt) return false;
         return true;
    }

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key() {
      return mKey;
    }
    
    BTreeCubeFirstDimensionSliceIterator<DataType> * clone() const {return new BTreeCubeFirstDimensionSliceIterator<DataType>(*this);}
  protected:

  
    vector<int> mKey;    
    DataType mCurrentValue;
    const BTreeCube<DataType> & mOwner;
    int  mAt;
    vector<int> mStart, mEnd;

};
#endif
