 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef CDBCube_H
#define CDBCube_H
#include <fstream>
#include "fileutil.h"
extern "C" {
#include <tinycdb.h> /* do not replace by <cdb.h> because this will clash againt the original cdb.h library!!! Let's think about this carefully*/
}
/* You can download TinyCDB from there: 
* http://www.corpit.ru/mjt/tinycdb/tinycdb-0.74.tar.gz  
* do make and make install.
*
* I could have used straight CDB, but I went for TinyCDB
* since the author claims to use less memory and the license
* is more compatible with GPL (CDB has a strange license!).
*
* [SEVERE Warning] It might be easier to look for
* a package called "tinycdb" before you try installing it on your own...
*/

#include "multidimensionalarray.h"

template<class DataType>
class CDBCubeIterator; 
template<class DataType>
class CDBCubeOverallIterator;  


using namespace std;
/**
* Very much like btreecube.h. However, uses a CDB hashtable instead of a btree.
*
* This is based around
* the tinycdb implementation
* http://www.corpit.ru/mjt/tinycdb.html
* of the CDB data structure. A CDB is a hashtable-like structure which
* is supposedly faster than a b-tree (maybe constant time though I doubt it).
* CDB is attributed to  Dan J. Bernstein.
* The trick with a CDB data structure is that you cannot delete a key!!!
*
* The keys are stored in the order they are entered and the iterator visits
* them in this order.
*
* One interesting part of the file is made of data records which is written in
* the following order "keylength, value length, key, value". It starts at byte
* 2048 (0800 in hex).
*
* TODO: we should write our own CDB optimize so we don't need store the value size. This would cut the size of the file.
*
 
*
* author: Daniel Lemire, July 2005
* 
*/
template<class DataType>
class CDBCube : public MultidimensionalArray<DataType> {

 
  public:

    /**
     * Create a cube from another cube. The data will be stored in 
     * a file (you must have write permissions to it, it will be created
     * if it doesn't exist).
     * TODO : here replace MultidimensionalArray<DataType> by something
     * more basic which would include iterators
     */ 
    CDBCube( const MultidimensionalArray<DataType>& , string filename = "CDBCubeTemp.bin") throw(std::ifstream::failure);
    CDBCube(vector<int> shape,MultidimensionalIterator<DataType>& mi, string filename = "CDBCubeTemp.bin") throw(std::ifstream::failure);
    /* if the file exists and you don't know its shape*/
    CDBCube( int NumberOfDimensions, string filename) throw(std::ifstream::failure);
    /* if the file exists and you know its shape*/
    CDBCube(vector<int> shape, string filename) throw(std::ifstream::failure);
    /* copy constructor */
    CDBCube(const CDBCube& bt): MultidimensionalArray<DataType>(bt), mFileName(bt.mFileName), mFD(bt.mFD), mValueSize(bt.mValueSize), mKeySize(bt.mKeySize), mDataEnd(bt.mDataEnd),mCDB(bt.mCDB){}
    
    virtual ~CDBCube();

    virtual int tempcode() const { return 3;}
    
    /**
    * return the name of the cube, typically a filename where the data
    * is stored, but could also be something like mysql:database/432:table2
    */
    string getName() const {return mFileName;}
   
    /**
     * In very rare cases, you might want to close the cube to 
     * release system resources. (Destructor will do it for you...)
     */
    void close();
   
    /**
     * If you ever happened to close the cube, you might need to open it.
     */
    void open() throw(std::ifstream::failure);
    /**
    * erase all 
    */
    void fillWithZeroes() {
             cerr<<"CDB ignoring fillWithZeroes!!!" <<endl;
    }
    
    /**
     * Query the cube. If the cell is not allocated, returns 0.
     */
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;

    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;

    /**
     * These calls are ignored. A warning is displayed.
     */
     void put(const vector<int>& pos, DataType value) {
             cerr<<"CDB ignoring put!!!" <<endl;
     }
    inline void put(const int pos[], DataType value){
             cerr<<"CDB ignoring put!!!" <<endl;
     }
    /**
    * load the content of this iterator! Notice that because
    * a CDBCube is constant, this will erase all previous content.
    */
    virtual void load(MultidimensionalIterator<DataType>&);



    typename MultidimensionalArray<DataType>::
      IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new CDBCubeIterator<DataType>(*this,start,end));
    }

    typename MultidimensionalArray<DataType>::IteratorPtr iterator(void) const {
 	return typename MultidimensionalArray<DataType>::
	  IteratorPtr(new CDBCubeOverallIterator<DataType>(*this));
     }
     
    void sync() {}

     int getNumberOfRecords() {return (mDataEnd-mDataBegin) /(mValueSize + mKeySize + 2* sizeof(uint *)) ;}

     enum{mDataBegin = 2048};// as per CDB definition
  protected:     
  

  private:
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape, string file) const  throw(std::ifstream::failure); 
    
    
    friend class CDBCubeIterator<DataType>; 
    friend class CDBCubeOverallIterator<DataType>;  
    string mFileName;
    int mFD;//file descriptor
    uint mValueSize, mKeySize,mDataEnd;
    cdb mCDB;
};

template<class DataType>
CDBCube<DataType>::CDBCube(const MultidimensionalArray<DataType>&bt, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(bt.getShape()),    mFileName(filename), mFD(0), mValueSize(sizeof(DataType)), mKeySize(sizeof(int)*this->mNumberOfDimensions), mDataEnd(0)  {
  std::auto_ptr<MultidimensionalIterator<DataType> > iter(bt.iterator());
  load(*iter);
}

template<class DataType>
CDBCube<DataType>::CDBCube(vector<int> shape,MultidimensionalIterator<DataType>& mi, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(shape),    mFileName(filename), mFD(0), mValueSize(sizeof(DataType)), mKeySize(sizeof(int)*this->mNumberOfDimensions), mDataEnd(0)  {
  load(mi);
}
template<class DataType>
void CDBCube<DataType>::load( MultidimensionalIterator<DataType>& it) {
  close();
  struct cdb_make cdbm;
  if (!( mFD = ::open(mFileName.c_str(), O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)) ) {
    cerr  << __FILE__ << ":"<< __LINE__ << " [CDBCube::load] I failed to open " << mFileName << endl;
    stringstream strs;
    strs << " [CDBCube::constructor] Failed to open file "<< mFileName<< ". I'm throwing an ifstream::failure!";    
    throw ifstream::failure(strs.str());
  }
  cdb_make_start(&cdbm, mFD); /* initialize structure */
  uint total = 0;
  //typename MultidimensionalArray<DataType>::IteratorPtr iter = bt.iterator();
  while (it.next()) {
          ++total;
          const void * key = &(it.key()[0]);
          DataType dt = it.value();
          const void * value = &dt;
          cdb_make_add(&cdbm, key, mKeySize,value, mValueSize);
  }
  mDataEnd = cdbm.cdb_dpos;
  assert(cdbm.cdb_rcnt == total);
  //cout << "CDB: I wrote "<< total<< " records!!! And I didn't crash yet!"<<endl;
  assert((uint) getNumberOfRecords()  == total);  
  cdb_make_finish(&cdbm);/* final stage - write indexes to CDB file */
  ::close(mFD);
  mFD= 0;
  open();// open for reading
}

template<class DataType>
CDBCube<DataType>::CDBCube(vector<int> shape, string filename) throw(std::ifstream::failure) :  MultidimensionalArray<DataType>(shape),    mFileName(filename), mFD(0), mValueSize(sizeof(DataType)), mKeySize(sizeof(int)*this->mNumberOfDimensions), mDataEnd(0)  {
  open();
}

template<class DataType>
CDBCube<DataType>::CDBCube(int NumberOfDimensions, string filename) throw(std::ifstream::failure) :
 MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)),    mFileName(filename), mFD(0), mValueSize(sizeof(DataType)), mKeySize(sizeof(int)*this->mNumberOfDimensions), mDataEnd(0) {
  open();

}


template<class DataType>
MultidimensionalArray<DataType>* CDBCube<DataType>::createNew(vector<int> shape, string file) const  throw(std::ifstream::failure){
        return new BTreeCube<DataType>(shape,file);// not a mistake!!! we couldn't return a cdbcube because it is immutable!!!
}



 
template<class DataType>
void CDBCube<DataType>::close() {
        if(mFD) {
                cdb_free(&this->mCDB);
                ::close(mFD);
                mFD = 0;
        }
}


template<class DataType>
CDBCube<DataType>::~CDBCube() {
    if(this->getCount() == 1) close();
} 
 
template<class DataType>
void CDBCube<DataType>::open() throw(ifstream::failure) {
  if(! FileUtil::fileExists(mFileName.c_str())) return;// the file will be loaded later! meanwhile, object is in a bad state,, FIXME
  close();
  //cout << "[CDB] Opening file "<< mFileName.c_str()<< " which has size = "<<FileUtil::getFileSize(mFileName.c_str())<< endl;
  if( FileUtil::getFileSize(mFileName.c_str())<2048) {
    cerr  << __FILE__ << ":"<< __LINE__ << " [CDB::open] This file is too small to be a CDB file: " << mFileName <<", size = "<< FileUtil::getFileSize(mFileName.c_str()) <<endl;
    stringstream strs;
    strs  << __FILE__ << ":"<< __LINE__ << " [CDB::open] This file is too small to be a CDB file: " << mFileName <<", size = "<< FileUtil::getFileSize(mFileName.c_str()) <<endl;
    throw ifstream::failure(strs.str());
  }
  if( ! (mFD = ::open(mFileName.c_str(), O_RDONLY))) {
    cerr  << __FILE__ << ":"<< __LINE__ << " [CDB::open] I failed to open " << mFileName << endl;
    stringstream strs;
    strs  << __FILE__ << ":"<< __LINE__ << " [CDB::open] I failed to open " << mFileName << endl;
    throw ifstream::failure(strs.str());
  }
  if(cdb_init(&this->mCDB, mFD)<0) {
    cerr  << __FILE__ << ":"<< __LINE__ << " [CDB::open] failed to init a cdb struct on  " << mFileName << " error code is " << errno<< " (see errno.h)"<< endl;
    stringstream strs;
    strs  << __FILE__ << ":"<< __LINE__ << " [CDB::open] failed to init a cdb struct on  " << mFileName << " error code is " << errno << " (see errno.h)" << endl;
    ::close(mFD);
    throw ifstream::failure(strs.str());
  }
  mDataEnd = this->mCDB.cdb_dend;
}


                

template<class DataType>
DataType CDBCube<DataType>::get(const vector<int>& pos) const {
        DataType val(0);
        cdb cdb(this->mCDB);
        if(cdb_find(&cdb, & pos[0], mKeySize) > 0) {
                assert(cdb_datalen(&cdb) == mValueSize);
                assert(cdb_keylen(&cdb) == mKeySize);
                cdb_readdata(&cdb, &val);
        } 
        return val;
}


template<class DataType>
DataType CDBCube<DataType>::get(const int pos[]) const {
        DataType val(0);
        cdb cdb(mCDB);        
        if(cdb_find(&cdb, & pos[0], mKeySize) > 0) {
                assert(cdb_datalen(&cdb) == mValueSize);
                assert(cdb_keylen(&cdb) == mKeySize);
                cdb_readdata(&cdb, &val);
        } 
        return val;
}

template<class DataType>
bool CDBCube<DataType>::undefinedAt(const int pos[]) const {
        cdb cdb(mCDB);
        return (cdb_find(&cdb, & pos[0], mKeySize) <= 0);
}



template<class DataType>
bool CDBCube<DataType>::undefinedAt(const vector<int>& pos) const {
        cdb cdb(mCDB);
        return (cdb_find(&cdb, & pos[0], mKeySize) <= 0);
}


/**
* Object cdbcube is completed
*/


/**************************************
* range iterator
* Optimized to go through a range (defined as the subcube
* spanned by any two cell). The last cell is "not included"
* so that you'd go through an array of length n by specifying
* 0 and n.
*
*************************************/

template <class DataType>
class CDBCubeIterator:public MultidimensionalIterator<DataType> {

  public: 
    CDBCubeIterator(const CDBCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
      MultidimensionalIterator<DataType>(),mStart(start), mEnd(end),mOwner(owner),  mKey(mStart),
      mCurrentValue(0),mPos(owner.mDataBegin-owner.mKeySize-owner.mValueSize), mCDB(mOwner.mCDB){
                  cdb_seqinit(&mPos,&mCDB);
      }

    CDBCubeIterator(const CDBCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      mStart(other.mStart), mEnd(other.mEnd), mOwner(other.mOwner), mKey(other.mKey), 
      mCurrentValue(other.mCurrentValue),mPos(other.mPos), mCDB(mOwner.mCDB)
    {}

    inline bool next() {
        do {
          if(!cdb_seqnext(&mPos,&mCDB)) return false;
          cdb_readdata(&mCDB, &mCurrentValue);
          cdb_readkey(&mCDB, &mKey[0]);
        } while(! isInRange());
	return true;  
    }    

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key()  {
      return mKey;
    }
    
    CDBCubeIterator<DataType> * clone() const {return new CDBCubeIterator<DataType>(*this);}
  protected:

  
    inline bool isInRange() {
            for(uint k = 0 ; k < mStart.size(); ++k) {
                    if((mKey[k] < mStart[k]) || (mEnd[k] > 0 && mKey[k]>= mEnd[k] )) {
                            return false;
                    }
            }
            return true;
    }
    vector<int> mStart, mEnd;
    const CDBCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    uint mPos;
    cdb mCDB;
};



/**************************************
* Overall iterator
* Optimized to go through the entire cube
*
*************************************/

template <class DataType>
class CDBCubeOverallIterator:public MultidimensionalIterator<DataType> {

  public: 
    CDBCubeOverallIterator(const CDBCube<DataType> & owner) :
      MultidimensionalIterator<DataType>(),mOwner(owner),  mKey(owner.mNumberOfDimensions,0),
      mCurrentValue(0),mPos(0), mCDB(mOwner.mCDB){
                  cdb_seqinit(&mPos,&mCDB);

      }

    CDBCubeOverallIterator(const CDBCubeOverallIterator<DataType> & other):
      MultidimensionalIterator<DataType>(), mOwner(other.mOwner), mKey(other.mKey), 
      mCurrentValue(other.mCurrentValue),mPos(other.mPos), mCDB(mOwner.mCDB)
    {}
    
    inline bool next() {
          if(!cdb_seqnext(&mPos,&mCDB)) return false;
          cdb_readdata(&mCDB, &mCurrentValue);
          cdb_readkey(&mCDB, &mKey[0]);
  	  return true;  
    }    

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key()  {
      return mKey;
    }
    
    CDBCubeOverallIterator<DataType> * clone() const {return new CDBCubeOverallIterator<DataType>(*this);}
  protected:

  
    const CDBCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    uint mPos;
    cdb mCDB;
};



#endif
