 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
 
 /*
 * TODO: Sparse chunks should probably be stored using a B-Tree. Can be achieved
 * using existing DBM libraries such as Berkeley DB. This means that knowing when
 * to switch  from sparse to dense might be more complex because of the B-tree
 * overhead. Also, iterators have to tweaked. 
 */

#ifndef HOLAPCUBE_H
#define HOLAPCUBE_H



#include <string>
#include <fstream>
#include <cassert>
#include <stdexcept>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include "bigarray.h"
// repeat after me: macros are evil and I shall never use them
#define HOLAP_PAIR_BUFFER 32

using namespace std;



template <class DataType>
class CubeIterator;


/***
 * THIS IS OBSELETE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * 
 * This class represent an HOLAP cube as defined by Kaser and Lemire and others.
 *
 * For simplicity, it is assumed that DataType(0) is symbolically equivalent
 * to "None", that is, no value is stored. If Dense storage is to be possible,
 * we need "None" to have some value, and 0 seems right. (Note that DataType
 * doesn't have to be a number.)
 *
 * We also have no way to "remove" values for the cube except to set them to
 * DataType(0).
 *
 * We have two in-memory data structures. One multidimensional array of 
 * bools to determine if a given chunk is dense or not, and one multidimensional
 * array of uints to keep track of how many cells are allocated in each
 * chunk.
 *
 * We initialize the object by providing a directory. Hopefully, this
 * code can support almost infinitely big date cubes!!!
 *
 *
 * author: Daniel Lemire, March 2004
 * 
 */
template<class DataType>
class HolapCube : public MultidimensionalArray<DataType> {

 
  public:

    /**
     * Create a cube with chunks of a given shape. The data will be stored in 
     * a directory (you must have write permissions to it, it will be created
     * if it doesn't exist). The basis is the basis of the b-tree the files
     * will be in. 4 or 5 appears like a reasonable choice, but the best choice
     * depends on how many chunks you have and your file system.
     */
    HolapCube(vector<int> shape, vector<int> chunkshape, string dirname, uint basis = 4);
    HolapCube(vector<int> shape, string dirname);
    
    virtual ~HolapCube();
    
    /**
    * return the name of the cube, typically a filename where the data
    * is stored, but could also be something like mysql:database/432:table2
    */
    string getName() const {return mDirName;}
    
    /**
     * In very rare cases, you might want to close the cube to 
     * release system resources. (Destructor will do it for you...)
     */
    void close();
   
    /**
     * If you ever happened to close the cube, you might need to open it.
     */
    void open();
    /**
    * erase all files
    */
    void fillWithZeroes();
    
    /**
     * Query the cube. If the cell is not allocated, returns 0.
     */
    DataType get( const vector<int>& pos) const;

    /**
     * put a value at a given cell. If the value is 0, then 
     * it should delete the cell, but in practice it doesn't (need packing
     * for sparse encoding, and no concept of deletion for dense coding).
     * It is "virtually" deleted (deleted in the count of allocated cells
     * for this chunk).
     */
    void put(const vector<int>& pos, DataType value);
    
    void sync() {}

    /**
     * This will iterator through all of the allocated cells
     * in the range from start (inclusively) to end (non-inclusively).
     * In theory, this should be really fast. You don't know in what
     * order the data will be traversed, but you know that each cell
     * in the range will be visited only once. An allocated cell is a
     * cell different from zero.
     */
    typename MultidimensionalArray<DataType>::IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
      return CubeIterator<DataType>(new CubeIterator<DataType>(*this,start,end));
    }

    /**
     * Return the shape in terms of number of chunks (see getNumberOfChunks).
     * Not sure why you'd want to do this.
     */
    vector<int> getChunkedShape() const;
    /**
     * Return how many chunks there are in the cube. Not sure why this 
     * is useful.
     */
    uint64 getNumberOfChunks() const;

    
  protected:
    /**
     * Gives the chunk coordinates for given cell coordinates (that is, where is the chunk?)
     */ 
    vector<int> getChunkCoordinates(vector<int> pos) const;


    /**
     * check in an in-memory array of booleans  to verify if the
     * chunk is dense. If it is, then we'll be able to locate the
     * information very quickly. If it isn't then life isn't easy.
     *
     * When you start, everything is dense, then after some time,
     * you will start seeing dense chunks appear.
     */
    bool isDense(const vector<int>& chunkcoordinates) const;

    /**
     * return how many cells are used in a given chunk
     */
    uint count(const vector<int>& chunkcoordinates) const;
   
    /**
     * One one cell was allocated. This automatically calls
     * makeItDense as needed.
     */
    void incrementCount(const vector<int>& chunkcoordinates);

    /**
     * One cell was deleted. This automatically calls makeItSparse as
     * needed.
     */
    void decrementCount(const vector<int>& chunkcoordinates);
    /**
     * given a chunk number, we derive a path
     */
    string getPath(uint64 chunknumber) const; 
  
        
    static vector<int> computeChunkedShape(vector<int> shape, const vector<int>& chunkshape);

  private:
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape, string file) const {return new HolapCube<DataType>(shape,file);}  
  
    vector<int> mChunkShape;
    
    std::string mDirName;

    /**
     * sign is dense/sparse, absolute value is
     * number of cells allocated, so that >=0 means sparse, < 0 means dense
     * which means that empty chunks have to be sparse
     */
    BigArray<int> mChunkDescriptor;
    
    uint mBasis;
    
    pair<uint, DataType> mPairBuffer[HOLAP_PAIR_BUFFER]; // some buffer for storing pairs... 32 is arbirary
 
    /**
     * Uglyness: computes the offset of the cell *inside* its chunk.
     */
    uint inChunkOffset(const vector<int>& pos, const vector<int> & chunkcoordinates) const;
  

    uint chunkVolume(const vector<int> & chunkcoordinates);
    
    bool shouldItBeDense(uint count, uint volume);

    void makeItDense(const vector<int> & chunkcoordinates);
    void makeItSparse(const vector<int> & chunkcoordinates);
};


template<class DataType>
HolapCube<DataType>::HolapCube(vector<int> shape, vector<int> chunkshape, 
    string dirname, uint basis) : MultidimensionalArray<DataType>(shape), mChunkShape(chunkshape), 
    mDirName(dirname),
    mChunkDescriptor(computeChunkedShape(mShape,mChunkShape),mDirName+"/chunkdescriptor.bin"),
    mBasis(basis)
   {}


template<class DataType>
HolapCube<DataType>::HolapCube(vector<int> shape,  
    string dirname) : MultidimensionalArray<DataType>(shape),  
    mDirName(dirname)
   {
   mBasis = 4;
   vector<int> mChunkShape(mShape);
   for(int k = 0;k < mNumberOfDimensions; ++k)
           mChunkShape[k] = mShape[k] > 4 ? mShape[k]/4 :mShape[k];
   mChunkDescriptor(computeChunkedShape(mShape,mChunkShape),mDirName+"/chunkdescriptor.bin");
   }

template<class DataType>
void HolapCube<DataType>::close() {
 mChunkDescriptor.close(); 
}
 
template<class DataType>
void HolapCube<DataType>::open() {
 mChunkDescriptor.open(); 
}
 
template<class DataType>
void HolapCube<DataType>::fillWithZeroes() {
  // I think I might go to hell for writing code like this but I could not find a way, 
  // in C to actually delete a bunch of files. 
  mChunkDescriptor.close();// not thread safe, but at least, I try to avoid kernel crashes
  ostringstream command;
  command << "rm -r -f " << mDirName<< "/[0-9]* "<< mDirName<< "/chunkdescriptor.bin";
  int ret = system(command.str().c_str());
  if(ret != 0) cerr << "Problem while removing files files in " << mDirName << endl;
  mChunkDescriptor.open();// open should make sure it recreates the file
}


template<class DataType>
HolapCube<DataType>::~HolapCube() {
  close();
} 

template<class DataType>
bool HolapCube<DataType>::isDense(const vector<int>& chunkcoordinates) const {
  return (mChunkDescriptor.get(chunkcoordinates) < 0);
} 

template<class DataType>
uint HolapCube<DataType>::count(const vector<int> & chunkcoordinates) const {
  if(vverbose) {
    cout << "[vverbose] count = " << abs(mChunkDescriptor.get(chunkcoordinates)) << endl;
  }
  return abs(mChunkDescriptor.get(chunkcoordinates));
}

template<class DataType>
void HolapCube<DataType>::incrementCount(const vector<int> & chunkcoordinates) {
  int oldvalue = mChunkDescriptor.get(chunkcoordinates);
  if(oldvalue < 0) mChunkDescriptor.put(chunkcoordinates,oldvalue - 1);
  else {
    if(shouldItBeDense(oldvalue + 1,chunkVolume(chunkcoordinates))) {
      mChunkDescriptor.put(chunkcoordinates, - oldvalue - 1);
      makeItDense(chunkcoordinates);
    } else
      mChunkDescriptor.put(chunkcoordinates,oldvalue + 1);
  }
  if(vverbose) {
    cout << "[vverbose] increment from " << oldvalue << " to " << mChunkDescriptor.get(chunkcoordinates) << endl;
  }
 }

template<class DataType>
void HolapCube<DataType>::decrementCount(const vector<int> & chunkcoordinates) {
  int oldvalue = mChunkDescriptor.get(chunkcoordinates);
  if(oldvalue < 0) { //
    if(! shouldItBeDense(oldvalue + 1, chunkVolume(chunkcoordinates) ) ) {
      mChunkDescriptor.put(chunkcoordinates,-oldvalue - 1);
      makeItSparse(chunkcoordinates);
    } else mChunkDescriptor.put(chunkcoordinates,oldvalue + 1); 
  } else mChunkDescriptor.put(chunkcoordinates,oldvalue - 1);
  if(vverbose) {
    cout << "[vverbose] decrement from " << oldvalue << " to " << mChunkDescriptor.get(chunkcoordinates) << endl;
  }
}

template<class DataType>
bool HolapCube<DataType>::shouldItBeDense(uint count, uint volume) {
  return count * sizeof(pair<uint,DataType>) >= volume * sizeof(DataType);
}

template<class DataType>
void HolapCube<DataType>::makeItDense(const vector<int> & chunkcoordinates) {
  if(verbose) cout << "[verbose] making it dense" << endl;
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  string path = getPath(chunk_offset);
  ifstream file( path.c_str(),std::ios::binary   ) ;
  if(!file.good()) {
        file.close();
        cerr << "Warning: file " << path << " doesn't exist, yet I'm asked to make it dense"<< endl; // great error handling
        cerr << "This is most certainly a bug as such storage would be highly inefficient." << endl;
        return;
  }
  string tmppath = path+".dense.temp";
  uint vol = chunkVolume(chunkcoordinates) * sizeof(DataType);
  FileUtil::fillFile(tmppath.c_str(), vol);
  uint actual_vol = FileUtil::getFileSize(tmppath.c_str());
  if(verbose) cout << " target vol = " << vol << " actual vol = " << actual_vol << endl;
  assert(actual_vol == vol);
  ofstream filetmp( tmppath.c_str(), std::ios::binary ); // not especially thread safe // nocreate is no longer allowed?
  uint number = count(chunkcoordinates);
  uint size_of_one_entry = sizeof(pair<uint,DataType>);
  while(number > 0) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        number -= entries_read;
        for(uint i = 0; i < entries_read; ++i) {
          filetmp.seekp(mPairBuffer[i].first * sizeof(DataType));
          filetmp.write((const char *) & mPairBuffer[i].second, sizeof(DataType));
        }
  }
  file.close();
  assert(filetmp.good());
  filetmp.close();
  if(vverbose) cout << " tmpfile size = " <<  FileUtil::getFileSize(tmppath.c_str()) << endl;
  FileUtil::copy(tmppath.c_str(),path.c_str());
  assert(FileUtil::getFileSize(path.c_str()) == vol);
  remove(tmppath.c_str());
}

template<class DataType>
void HolapCube<DataType>::makeItSparse(const vector<int> & chunkcoordinates) {
  if(verbose) cout << "[verbose] making it parse" << endl;
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  string path = getPath(chunk_offset);
  ifstream file( path.c_str(),std::ios::binary ) ;
  string tmppath = path+".sparse.temp";
  if(!file.good()) {
        file.close();
        cerr << "Warning: file " << path << " doesn't exist!"<< endl; // great error handling
        cerr << "I'm asked to make it sparse. Why wasn't it sparse before?" << endl;
        cerr << "This is probably a (harmless?) bug." << endl;
        return;
  }
  ofstream filetmp( tmppath.c_str(),std::ios::binary | std::ios::trunc ) ;
  uint buffer_size = 32;
  DataType buffer[buffer_size];
  uint actualread = 0;
  uint total_components_read = 0;
  while((actualread = file.read((char *)& buffer,buffer_size).gcount())>0) {
    uint components_read = actualread / sizeof(DataType);
    for(uint i = 0; i < components_read; ++i) {
      if(buffer[i] != 0) {
        uint component_position = total_components_read + i;
        pair<uint,DataType> entry(component_position, buffer[i]);
        filetmp.write((const char *)&entry,sizeof(entry));
      }
    }
    total_components_read += components_read;
  }
  file.close();
  assert(filetmp.good());
  filetmp.close();
  FileUtil::copy(tmppath.c_str(),path.c_str());
  remove(tmppath.c_str());
}
 
template<class DataType>
vector<int> HolapCube<DataType>::getChunkedShape() const {
  return mChunkDescriptor.getShape();
} 

template<class DataType>
uint64 HolapCube<DataType>:: getNumberOfChunks() const {
  return mChunkDescriptor.getVolume();
}

template<class DataType>
vector<int> HolapCube<DataType>::getChunkCoordinates(vector<int> pos) const {
  for(uint i = 0; i < pos.size(); ++i) pos[i] /= mChunkShape[i];
  return pos;
}

   

template<class DataType>
DataType HolapCube<DataType>::get(const vector<int>& pos) const {
  /**
   * The current code opens and close a file for each query. That's
   * probably terrible from a performance point of view so one would want
   * to pool the fstream some place.
   */
  vector<int> chunkcoordinates = getChunkCoordinates(pos);
  bool is_dense = isDense(chunkcoordinates);
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  uint64 offset_in_chunk =  inChunkOffset(pos,chunkcoordinates); 
  string path = getPath(chunk_offset);
  if(verbose) cout << "[verbose] Doing a get on file "<< path << endl;
  ifstream file( path.c_str(),std::ios::binary | std::ios::in) ;
  if(!file.good()) {
        file.close();
        if(verbose) 
          cout << "[verbose] File " << path << " doesn't exist yet"<< endl; // great error handling
        return DataType(0);
  }
  DataType value(0);
  if(is_dense) {
     uint local_offset = offset_in_chunk * sizeof(DataType); 
     if(verbose) cout << "[verbose] file is dense, going to position " << local_offset << endl;
     file.seekg(local_offset);
     file.read((char*) & value,sizeof(DataType));
     assert(file.gcount() == sizeof(DataType));
  } else {
      uint number = count(chunkcoordinates);
      if(verbose) cout << "[verbose] file is sparse, there are " << number << " entries." << endl;
      // in theory, we could read the pairs one by one, but in practice, this
      // would be truly bad, so I create a buffer
      uint size_of_one_entry = sizeof(pair<uint,DataType>);
      while(number > 0) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        if(vverbose) cout << "[vverbose] trying to read " << toberead << " bytes. " << endl;
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        if(vverbose) cout << "[vverbose]  read " << file.gcount() << " bytes. " << endl;
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        for(uint i = 0; i < entries_read; ++i) {
          if(mPairBuffer[i].first == offset_in_chunk) {
            if(verbose) cout << "[verbose] I found a match with "<< (number - i) << " left to check." <<endl;
            file.close();
            return mPairBuffer[i].second;
          }
        }
        number -= entries_read;
      }
      if(verbose) cout << "[verbose] A match was not found. I will return 0." <<endl;
  }
  assert(file.good());
  file.close();
  return value;
}


template<class DataType>
void HolapCube<DataType>::put(const vector<int>& pos, DataType value) {
  // don't forget to increment counter
  vector<int> chunkcoordinates = getChunkCoordinates(pos);
  bool is_dense = isDense(chunkcoordinates);
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  uint64 offset_in_chunk =  inChunkOffset(pos,chunkcoordinates); 
  string path = getPath(chunk_offset);
  if(verbose) cout << "[verbose] Doing a put on file "<< path << endl;
  fstream file( path.c_str(),std::ios::binary | std::ios::out | std::ios::in  ) ;
  // if the file doesn't exist, file opening will fail
  if(!file.good()) {
    file.clear();
    file.open(path.c_str(),std::ios::binary | std::ios::out | std::ios::in | std::ios::trunc ) ;
    // should do error handling
    if(!file.good()) 
      cerr <<"[error] I can't open " << path << " for a put. " << endl;
    assert(file.good());
  }
  DataType oldvalue (0);
  if(is_dense) {// data is densely coded
     uint local_offset = offset_in_chunk * sizeof(DataType); 
     if(verbose) cout << "[verbose] file is dense, going to position " << local_offset << endl;
     file.seekg(local_offset);
     file.read((char *) & oldvalue, sizeof(DataType));// we need this for the count
     file.seekp(local_offset);
     file.write((const char*) & value,sizeof(DataType));
  } else {// data is sparsely coded
      uint number = count(chunkcoordinates);
      if(verbose) cout << "[verbose] file is sparse, there are " << number << " entries." << endl;
      // in theory, we could read the pairs one by one, but in practice, this
      // would be truly bad, so I create a buffer called mPairBuffer
      uint size_of_one_entry = sizeof(pair<uint,DataType>);
      uint64 writepos = 0;
      bool matchFound = false;
      // ugly code follows
      while((number > 0) && (! matchFound)) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        if(vverbose) 
          cout << "[vverbose] file.gcount() = " << file.gcount()  << " toberead = " << toberead << endl;
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        for(uint i = 0; i < entries_read; ++i) {
          writepos += size_of_one_entry;// new line here
          if(mPairBuffer[i].first == offset_in_chunk) {
            if(verbose) cout << "[verbose] I found a match with "<< (number - i) << " left to check." <<endl;
            if(vverbose) cout << "[vverbose] i = "<< i << " number = " << number 
              << " entries_read = " << entries_read << " size_of_one_entry = " 
              << size_of_one_entry << endl;
            oldvalue = mPairBuffer[i].second;
            if( vverbose) 
              cout << file.tellg() << " - ( " << entries_read << " - " << i << " ) * " 
                << size_of_one_entry << endl;
            
            //writepos = file.tellg() -  (entries_read - i) *size_of_one_entry;
            break;
          }
        }
        number -= entries_read;
      }
      if(! matchFound) {
        //writepos = file.tellg();
        if(verbose) 
          cout << "[verbose] A match was not found. I will write the value at the end :"<< writepos <<endl;
      } else {
        if(verbose) cout << "[verbose] Match was found at position : " << writepos << endl;
      }
      file.seekp(writepos);
      pair<uint,DataType> new_entry(offset_in_chunk,value);
      file.write((char *) & new_entry, sizeof(new_entry));
  }
  assert(file.good());
  file.close();
  if(verbose) cout <<" oldvalue = " << oldvalue << " value = " << value << endl;
  if((oldvalue == 0 ) && (value != 0)) {
    if(verbose) cout << " incrementing count " << endl;
    incrementCount(chunkcoordinates);
  }
  else if ((oldvalue != 0) && (value == 0)) {    
    if(verbose) cout << " decrementing count " << endl;
    decrementCount(chunkcoordinates);      
  }
}

template<class DataType>
string HolapCube<DataType>::getPath(uint64 chunknumber) const {
  uint lastdigit = chunknumber % mBasis;
  uint64 whatremains = chunknumber / mBasis;
  ostringstream ost;
  ost << mDirName;
  while(whatremains > 0) {
    ost << "/" << (whatremains % mBasis);
    whatremains /= mBasis;
  }
  if(! FileUtil::fileExists(ost.str().c_str()) ) { // if directories don't exist, create them
    if( FileUtil::mkdir_parents(ost.str().c_str()) ) // make sure the path exists!!!
      cerr << " I couldn't create the proper directory. This program will fail. " 
        << endl;// great error handling  
  }
  ost << "/" << lastdigit << ".bin";
  return ost.str();
}


/**
 * Uglyness: computes the offset of the cell *inside* its chunk.
 */
template<class DataType>
uint HolapCube<DataType>::inChunkOffset(const vector<int>& pos, const vector<int> & chunkcoordinates) const {
  uint offset = 0;
  uint base = 1;
  for (uint i = 0; i < pos.size(); ++i) {
    offset += (pos[i] % mChunkShape[i] ) * base;
    base *=  min((chunkcoordinates[i]+1)*mChunkShape[i] , mShape[i] ) 
      - chunkcoordinates[i] * mChunkShape[i]; 
    // compute the "width" of the chunk
    // most time it will just b mChunkShape[i], but we have the "edges" to worry about
  }
  return offset;
}

template<class DataType>
uint HolapCube<DataType>::chunkVolume(const vector<int> & chunkcoordinates) {
  uint volume = 1;
  for (uint i = 0; i < chunkcoordinates.size() ; ++i) {
    volume *= min((chunkcoordinates[i]+1)*mChunkShape[i] , mShape[i] ) 
      - chunkcoordinates[i] * mChunkShape[i]; 
  }
  return volume;
}

template<class DataType>
vector<int> HolapCube<DataType>::computeChunkedShape(vector<int> shape, const vector<int>& chunkshape) {
  for (uint i = 0; i < shape.size(); ++i) {
    shape[i] = shape[i] / chunkshape[i] + (shape[i] % chunkshape[i] > 0 ? 1 : 0);
  }
  return shape;  
} 



/************************************************************
 * We are done with the holap cube class.
 *
 ***********************************************************/

/**
 * This is a suboptimal implementation of the iterator
 * over an Holap cube. What should be done is that the
 * iterator would run differently over dense and sparse
 * chunks and would most definitively not call the get
 * method directly on the cube. That's waaaaaayyyy too slow.
 * But for now, I'm prototyping. 
 *
 * FIX ME. MAKE ME FASTER.
 */
template <class DataType>
class CubeIterator: public MultidimensionalIterator<DataType> {

  public: 

    CubeIterator(const HolapCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
      mStart(start), mEnd(end),mIndex(start),
      mCurrentValue(0),mOwner(owner){
        mIndex[0] -= 1;// start right before the area
    }

    CubeIterator(const CubeIterator<DataType> & other):
      mStart(other.mStart), mEnd(other.mEnd), mIndex(other.mIndex), 
      mCurrentValue(other.mCurrentValue), mOwner(other.mOwner)
    {}
    /**
     * This is where the code needs to get smarter.
     */
    inline bool next() {
      if(! increment ()) return false;
      while((mCurrentValue = mOwner.get(mIndex)) == 0) 
        if(! increment ()) return false;
      return true;
    }

    DataType value() const {
      return mCurrentValue;
    }

    vector<int> key() const {
      return mIndex;
    }
    
  protected:
     inline bool increment() {
      for(uint dim = 0; dim < mIndex.size(); ++dim) {
        if(++mIndex[dim] >= mEnd[dim]) {
          mIndex[dim] = mStart[dim];
        } else {
          return true;
        }
      }
      return false;
    }   

    bool slow_increment(uint dim = 0) {
      if(++mIndex[dim] >= mEnd[dim]) {
        if(dim == mIndex.size() - 1) return false;
        mIndex[dim] = mStart[dim];
        return increment(++dim);
      }
      //cout << "increment to ";
      //for(uint k =0; k < mIndex.size() ; ++k)
      //  cout << mIndex[k] << " ";
      //cout << endl;
      return true;
    }

      
    vector<int> mStart, mEnd, mIndex;
    DataType mCurrentValue;
    const HolapCube<DataType> & mOwner;
};





#endif

