 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef INTERNALBIGARRAY_H
#define INTERNALBIGARRAY_H


#include "bigarray.h"




using namespace std;
 

template<class DataType>
class InternalOverallIterator; 
/***
 * This class represent a very large array stored using internal memory.
 *
 * author: Daniel Lemire, March 2005
 * 
 */
template<class DataType>
class InternalBigArray :  public MultidimensionalArray<DataType> {
        public:
                InternalBigArray(vector<int> shape) throw(std::bad_alloc);
                InternalBigArray(int shape[], int NumberOfDimensions) throw(std::bad_alloc);
                
                InternalBigArray(const InternalBigArray& bt): MultidimensionalArray<DataType>(bt), mData(bt.mData) {  }


                virtual ~InternalBigArray() {if(this->getCount() == 1) close();}

                virtual void fillWithZeroes();

                /**
                * return the name of the cube, typically a filename where the data
                * is stored, but could also be something like mysql:database/432:table2
                 */
                string getName() const {return "RAM";}
                // if you manually closed the mapping, you can reopen it
                virtual void open() throw(std::bad_alloc);
                
                // if you want, you can manually close the mapping
                virtual void close();
                
                inline DataType get(const int  pos[]) const;
                inline DataType get(const vector<int>&  pos) const;

		// undefinedAt should be based on a auxiliary bitmap
		inline bool undefinedAt(const int pos[]) const { return false;} // these are broken -ofk
		inline bool undefinedAt( const vector<int> & pos) const { return false;}


                // add delta and return new value
                inline DataType increment(const vector<int>&  pos, DataType delta) const;
                inline DataType increment(const int  pos[], DataType delta) const;
 
                inline void put(const vector<int>& pos, DataType value);
                inline void put(const int  pos[], DataType value);

                 typename MultidimensionalArray<DataType>::IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
                        return typename MultidimensionalArray<DataType>::IteratorPtr(new BigArrayIterator<DataType>(*this,start,end));
                 }
                 
                 
                 typename MultidimensionalArray<DataType>::IteratorPtr iterator() const {
                           return typename MultidimensionalArray<DataType>::IteratorPtr(new InternalOverallIterator<DataType>(*this));
                 }   

                 void sync() {}// does nothing
        private:

             virtual MultidimensionalArray<DataType>* createNew(vector<int> shape, string file)  const throw(std::bad_alloc, std::ifstream::failure);// {return new InternalBigArray<DataType>(shape);}  
             
             DataType *mData;// pointer to data
             inline DataType fastget(const int64  pos) const {
                        return mData[pos];
             }
                
             friend class InternalOverallIterator<DataType>;
             
};


template <class DataType>
InternalBigArray<DataType>::InternalBigArray(vector<int> Shape) throw(std::bad_alloc)
  : MultidimensionalArray<DataType>(Shape),mData(0){
          open();
}
template <class DataType>
InternalBigArray<DataType>::InternalBigArray(int Shape[], int NumberOfDimensions) throw(std::bad_alloc)
  : MultidimensionalArray<DataType>(Shape, NumberOfDimensions),mData(0){
          open();
}
template <class DataType>
MultidimensionalArray<DataType>* InternalBigArray<DataType>::createNew(vector<int> shape, string file)  const throw(std::bad_alloc, std::ifstream::failure) {return new InternalBigArray<DataType>(shape);}

template <class DataType>
void InternalBigArray<DataType>::open() throw(std::bad_alloc) {

  close();
  mData = new DataType[this->getVolume()];
}



template <class DataType>
void InternalBigArray<DataType>::close() {
  if (mData != NULL)  {
    delete[] mData;
    mData = NULL;
  }
}


template <class DataType>
void InternalBigArray<DataType>::fillWithZeroes()
{
  //fill memory with zeroes and sync file and memory.
  for (uint64 index = 0; index < this->getVolume(); index++)
    *(mData + index) = 0;
}




template <class DataType>
DataType InternalBigArray<DataType>::get(const vector<int>& pos) const {
  uint64 offset = this->computeOffset(pos);
  //cout << " offset = "<< offset << endl;
  const DataType answer = mData[offset];
  //cout << " done " <<endl;
  return answer;
}

template <class DataType>
DataType InternalBigArray<DataType>::get(const int pos[]) const {
  uint64 offset = this->computeOffset(pos);
  //cout << " offset = "<< offset << endl;
  const DataType answer = mData[offset];
  //cout << " done " <<endl;
  return answer;
}

template <class DataType>
DataType InternalBigArray<DataType>::increment(const vector<int>& pos, DataType delta) const {
  uint64 offset = this->computeOffset(pos);
  const DataType answer = (mData[offset]+=delta);
  return answer;
}

template <class DataType>
DataType InternalBigArray<DataType>::increment(const int pos[], DataType delta) const {
  uint64 offset = this->computeOffset(pos);
  const DataType answer = (mData[offset]+=delta);
  return answer;
}

template <class DataType>
void InternalBigArray<DataType>::put(const vector<int>& pos, DataType value) {
  uint64 offset = this->computeOffset(pos);
  *(mData + offset) = value;
}

template <class DataType>
void InternalBigArray<DataType>::put(const int pos[], DataType value) {
  uint64 offset = this->computeOffset(pos);
  *(mData + offset) = value;
}



template <class DataType>
class InternalOverallIterator:public MultidimensionalIterator<DataType> {

  public: 

    InternalOverallIterator(const InternalBigArray<DataType> & owner) : MultidimensionalIterator<DataType>(), mOwner(owner), mVolume(mOwner.getVolume()),mIndex(-1), mCurrentValue(0),mCurrentPos(mOwner.getShape().size()){

    }

    InternalOverallIterator(const InternalOverallIterator<DataType> & other):MultidimensionalIterator<DataType>(),mOwner(other.mOwner), mVolume(other.mVolume),mIndex(other.mIndex), mCurrentValue(other.mCurrentValue), mCurrentPos(other.mCurrentPos)
    {}

    inline bool next() {
      if(++mIndex >= mVolume) return false;
      while((mCurrentValue = mOwner.fastget(mIndex)) == 0) 
        if(++mIndex >= mVolume) return false;
      return true;
    }

    inline DataType value() const {
      return mCurrentValue;
    }

    inline vector<int>& key() {
      mOwner.reverseoffset(mIndex,mCurrentPos);
      return mCurrentPos;
    }
    
    InternalOverallIterator<DataType> * clone() const {return new InternalOverallIterator<DataType>(*this);}
    
  protected:
    
    const InternalBigArray<DataType> & mOwner;
    const int64 mVolume;
    int64 mIndex;
    DataType mCurrentValue;
    vector<int> mCurrentPos;


};



#endif
