 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef MULTIDIMENSIONALARRAY_H
#define MULTIDIMENSIONALARRAY_H

#include <vector>
#include <queue>
#include <stdexcept>
#include <sstream>
#include <memory>
#include <typeinfo>
#include <string>
#include "agmap.h"
#include "olap_operator.h"
#include "fileutil.h"

using namespace std;

template <class DataType>
class MultidimensionalIterator;
template <class DataType>
class IcebergMultidimensionalIterator;
template <class DataType>
class DeleteDimensionIterator;
template <class DataType>
class PrependDimensionIterator;
template <class DataType>
class GlueIterator;

template <class DataType>
class BTreeCube;
template <class DataType>
class BigArray;
template <class DataType>
class CDBCube;

template <class DataType>
class CubeSet;
template <class DataType>
class IteratorSet;

template <class DataType>
class MultidimensionalArray {
  public:
      // next value should be very large, but not so large as to exceed 32 bits int limit, 
                     
      enum {CUBETYPE_SAME=0, CUBETYPE_BTREECUBE, CUBETYPE_BIGARRAY,CUBETYPE_CDBCUBE,INVALID_SIZE=2147483647 };    
      typedef MultidimensionalIterator<DataType> * IteratorPtr;
      typedef MultidimensionalArray<DataType> * MultidimensionalArrayPtr;

      /**
      * By convention, INVALID_SIZE which is equal to 2147483647 means you don't know what the dimension is!
      */
      MultidimensionalArray(vector<int> shape) ; 
      MultidimensionalArray(int shape[], int NumberOfDimensions);       
      MultidimensionalArray(const MultidimensionalArray& ma) :
	mShape(ma.mShape),mNumberOfDimensions(ma.mNumberOfDimensions),mCount(ma.mCount),
	mVolume(ma.mVolume), mOffsetBase(ma.mOffsetBase) {
                *mCount +=1;
		//		cout << "copy constructor for Multidim array (an abstract class!  Hope this is not being used alone) " << endl;
      }
      virtual ~MultidimensionalArray();//{--mCount;}
      // how many compoents are there in total?
      // because this is called often, the result is precomputed
      inline uint64 getVolume() const { return mVolume; }
      inline string cubeType() const {return string(typeid(*this).name());}
      
      /**
      * return the name of the cube, typically a filename where the data
      * is stored, but could also be something like mysql:database/432:table2
      */
      virtual string getName() const = 0;
      inline vector<int> getShape() const {return mShape;}
    /**
     * Query the cube. If the cell is not allocated, returns 0.
     */
    virtual DataType get( const vector<int>& pos) const = 0;
    virtual DataType get( const int pos[]) const = 0;
    
    // in case we need to know "real" zeros from "undefined".  Design is broken if we cannot tell!
    virtual bool  undefinedAt( const int pos[]) const = 0;
    virtual bool  undefinedAt( const vector<int>& pos) const = 0;

    virtual int tempcode() const { return 1;}
    
    /**
    * overwrite this if getName() doesn't return a file modified each time
    * data is changed.
    */
    virtual int64 lastModification() {
        return FileUtil::lastModification(  getName().c_str()  );
    }
    
    virtual void close() = 0;
    
    /**
    * overwrite this if the data is stored in many files
    */
    virtual bool deleteOnDisk() {
            close();
            int result = FileUtil::unlink(  getName().c_str()  );
            return (result == 0);
    }
    
    virtual void increment(const int pos[]) { put(pos, get(pos)+1);   }
    virtual void increment(const vector<int>& pos) { put(pos, get(pos)+1);   }
    bool hasValidShape() const {
          return mShape[0]!=INVALID_SIZE;
    }

    /**
     * put a value at a given cell. If the value is 0, then 
     * it should delete the cell.
     *
     * OFK: I think we need to get away from the 0==undefined mindset...
     * Daniel: Ok. No strong feeling about it, but we need to add a delete 
     * to the API.
     */
    virtual void put(const vector<int>& pos, DataType value) = 0;
    virtual void put(const int pos[], DataType value) = 0;
      // for each cell returns a unique number, starting at zero 
      // all the way to the total number of cells
      inline uint64 computeOffset (const vector<int>& idx) const;
      inline uint64 computeOffset (const int idx[]) const;

      virtual void fillWithZeroes(void) = 0;
      
      /**
      * This is the range iterator.
      */
      virtual IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const  = 0;
      
    /**
    * This is a convenience range iterator method.
    */
    IteratorPtr iterator(const int start[], const int end[]) const {
              vector<int> s(mNumberOfDimensions), e(mNumberOfDimensions);
              for(uint i = 0; i < mNumberOfDimensions ; ++i) {
                      s[i] = start[i];
                      e[i] = end[i];
              }
              return iterator(s,e);                     
    }

    virtual IteratorPtr iterator() const {
              vector<int> s(mNumberOfDimensions,0);
              return iterator(s,mShape);                    
    }
    /**
    * load the content of this iterator!
    */
    virtual void load(MultidimensionalIterator<DataType>&);
    /**
    * You should try to derive sliceIterator with something faster when possible.
    */
    virtual IteratorPtr sliceIterator( int dim, int slice_at) const {
      vector<int> s(mNumberOfDimensions,0);  s[dim]= (int) slice_at;
      vector<int> e(mShape);  e[dim] = 1 + (int) slice_at;
      return iterator(s,e);
    }

    virtual void printData() const {
      for (std::auto_ptr<MultidimensionalIterator<DataType> >  it(iterator()); it->next(); ) {
	vector<int> where = it->key();
	cout << it->value() << " at ";
	for(uint i = 0; i < mNumberOfDimensions; ++i)
	  cout << where[i] << " ";
	cout << endl;
     }
    }
    
    /*    virtual MultidimensionalArrayPtr partial_rollup( vector<AgMap> agmaps, const string fname) const; */
 
    
    /* will die soon? */   
 virtual MultidimensionalArrayPtr  partial_rollup(const InvertedAgMap map[], const string fname, const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument);

    virtual MultidimensionalArrayPtr full_rollup( const vector<int>& dimsToKill, const string fname,const int CUBETYPE=CUBETYPE_SAME) const  throw(std::out_of_range, std::ifstream::failure, std::bad_alloc, std::invalid_argument);

    
    // convenience method
    virtual MultidimensionalArrayPtr full_rollup( int dimsToKill[], int numberofdimstokill, const string fname,const int CUBETYPE=CUBETYPE_SAME) const  throw(std::out_of_range,std::ifstream::failure, std::bad_alloc, std::invalid_argument) {
            vector<int> dimsToKillVec(numberofdimstokill);
            for(int k = 0; k < numberofdimstokill;++k) 
                    dimsToKillVec[k] = dimsToKill[k];
            return full_rollup( dimsToKillVec,  fname);
    }

    virtual MultidimensionalArrayPtr map_pointwise( const binary_olap_operator<DataType> &op, 
						    const MultidimensionalArray & operand2,
						    const string fname, const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument);

    MultidimensionalArrayPtr map_pointwise( const char opname, const MultidimensionalArray &op2, 
					    const string fname,const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument){
      binary_olap_operator<DataType> & op = binary_olap_operator_factory<DataType>::makeOperator(opname);
      MultidimensionalArrayPtr res = map_pointwise( op, op2, fname,CUBETYPE);
      return res;
    }

    virtual MultidimensionalArrayPtr map_1dvectors( const vector_olap_operator<DataType> &op, 
						    const MultidimensionalArray & op2,
						    const int vector_length,
						    const int component_dim,
						    const int component_dim2,
						    const int pre_threshold,
						    const int post_threshold,
						    const string fname,const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument);

    MultidimensionalArrayPtr map_1dvectors( const char opname, const MultidimensionalArray &op2,
					    const int vector_length,
					    const int component_dim,
					    const int component_dim2,
					    const int pre_threshold, const int post_threshold,
					    const string fname,const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument) {
      vector_olap_operator<DataType> & op = vector_olap_operator_factory<DataType>::makeOperator(opname);
      MultidimensionalArrayPtr res = map_1dvectors( op, op2, vector_length, component_dim, component_dim2, 
						    pre_threshold, post_threshold, fname,CUBETYPE);
      return res;
    }


    virtual MultidimensionalArrayPtr slice( const uint dim, const uint slice_at,
					    const string fname,const int CUBETYPE=CUBETYPE_SAME) const  throw(std::out_of_range,std::ifstream::failure, std::bad_alloc);

    virtual DataType topN_frequency(const uint n) const throw(std::out_of_range);  // order statistics

    virtual DataType findMaxVal() const throw(std::out_of_range);  // specialized version of topN_frequency(1);
    
    virtual int getNumberOfRecords();

    virtual MultidimensionalArrayPtr iceberg( const DataType threshold,  // materialize iceberg
					      const string fname,const int CUBETYPE=CUBETYPE_SAME) const throw(std::ifstream::failure, std::bad_alloc);
      /**
      * convenience method
      */

     /** this is the reverse of slice. (If we had a form of slice that returned a list of all slices.)
           All cubes must have same shape, result has one more dimension (the glue dimension) */

     virtual MultidimensionalArrayPtr
       glue( CubeSet<DataType> others, const string fname, const int CUBETYPE=CUBETYPE_SAME) const 
     throw(std::ifstream::failure, std::bad_alloc,std::invalid_argument);



       virtual MultidimensionalArrayPtr collapse( CubeSet<DataType> others,  
						  const vector<int> & attValueStarting, 
						  const int dim, const string fname, 
						  const int CUBETYPE=CUBETYPE_SAME) const 
	 throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument);
       
       /** convenience method */
       MultidimensionalArrayPtr collapse( CubeSet<DataType> others,  
					  int attValueStarting[], const int dim, 
					  const string fname, const int CUBETYPE=CUBETYPE_SAME) const 
	 throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument) {
               vector<int> vecattValueStarting(others.size() + 1,0);
               for(int k = 0; k < others.size() + 1; ++k) vecattValueStarting[k] = attValueStarting[k];
               return collapse(others, vecattValueStarting, dim,fname,CUBETYPE);
       }
       
    virtual MultidimensionalArrayPtr prepend_zero_coordinate( const string fname, const int CUBETYPE=CUBETYPE_SAME) const 
      throw(std::ifstream::failure, std::bad_alloc);
      
    
    /**
    * sync data with underlying storage
    */
    virtual void sync()=0;
    
    /**
    * map in the other direction, useful for translating iterators on keys
    * to iterators on tuples.
    * Daniel: should check whether this is in use
    */
    inline void reverseoffset(uint64 index, vector<int>& answer) const {
            for(uint i = 0; i < mNumberOfDimensions - 1; ++i ) {
                    const uint64 indexmod = index % mOffsetBase[i + 1]; 
                    answer[i] = indexmod / mOffsetBase[i]; 
                    index -= indexmod ;
            }            
	    answer[mNumberOfDimensions-1] = index/ mOffsetBase[mNumberOfDimensions - 1];
    }

    int getCountTemp() const { return getCount();} // temporary, for debugging
   
  protected:
      vector<int> mShape; 
      uint mNumberOfDimensions;
      
      int getCount() const {return *mCount;}// how many instance of this cube exist?

  private :
      int * mCount; // how many instance of this cube exist?
      enum{verbose = 100, verbosepartialrollup=true};
      // probabilistic rounding (of positive numbers) is required only integral types.  No-op for others
      // no doubt there is some way to use template overload resolution to encode 
      //  "DTypes except float and double are noop, rest use probabilistic rounding"
      // it needs to be pretty fast.
      inline double maybe_round(double d, double overloadHorror) const { return d;}
      // users of short and char will pay a conversion cost
      inline int maybe_round(double d, int overloadHorror) const {
	double frac = d - floor(d);  // hey, wasn't there a library fcn for this?
	if (frac * RAND_MAX > rand()) return 1 + int(d); 
	else return int(d);
      }

      virtual MultidimensionalArray<DataType> * createNew(vector<int> shape, string file) const throw(std::ifstream::failure, std::bad_alloc) = 0;  //
      //this is the virtual constructor pattern

      MultidimensionalArray<DataType> * createNewByType(vector<int> shape, string file, const int CUBETYPE) const; 
      
      uint64 mVolume; // precomputed volume for fast computations
      // precompute how many components are there in total?
      void precomputeVolume();
      vector<uint64> mOffsetBase;//use in computeOffset for fast computations
       
};



/****************************************************************
****************************************************************
****************************************************************
* Now that the class has been declared, we define the methods.
**************************************************************
****************************************************************
****************************************************************/



template <class DataType>
MultidimensionalArray<DataType>::MultidimensionalArray(vector<int> shape) : 
	mShape(shape),mNumberOfDimensions(mShape.size()), mCount(new int(1)),mOffsetBase(mShape.size(),1) {
        for(uint i = 1; i < mOffsetBase.size(); ++i ) mOffsetBase[i] = mOffsetBase[i-1] * mShape[i - 1];
        precomputeVolume(); 
      }

template <class DataType>
MultidimensionalArray<DataType>::MultidimensionalArray(int shape[], int NumberOfDimensions) :
        mShape(NumberOfDimensions),mNumberOfDimensions(NumberOfDimensions), mCount(new int(1)),mOffsetBase(mNumberOfDimensions,1) {
        for(uint i =0; i < mNumberOfDimensions;++i) mShape[i] = shape[i];
        for(uint i = 1; i < mOffsetBase.size(); ++i ) mOffsetBase[i] = mOffsetBase[i-1] * mShape[i - 1];
        precomputeVolume(); 
      }
template <class DataType>
MultidimensionalArray<DataType>::~MultidimensionalArray(){
        *mCount-=1;
        assert(getCount() >=0);
        if(*mCount==0) delete mCount;
        mCount = 0;
}
 
/*template <class DataType>
MultidimensionalArray<DataType> & MultidimensionalArray<DataType>::operator=(const MultidimensionalArray<DataType>& ma) {
        cerr << "assignment operator not implemented. You are going to hell!" << endl;
        return *this;
}
*/


template <class DataType>
uint64 MultidimensionalArray<DataType>::computeOffset(const vector<int>& idx) const  {
                  //notice here that idx doesn't necessarily lie in the cube, could be outside since we may iterate up to, but not including a given point outside the allowed values.
                  uint64 offset = 0; 
		  
		  //		  cout << "compute offset ";
                  for(uint i = 0; i < mNumberOfDimensions; ++i ) {
		    //  cout << idx[i] << " * " << mOffsetBase[i] << "   ";
		    offset += idx[i] * mOffsetBase[i];
		  }
                  //if (offset >= mVolume) throw std::out_of_range("hashtable_computeoffset");
		  // cout << offset << endl;
                  return offset;
}

template <class DataType>
uint64 MultidimensionalArray<DataType>::computeOffset(const int idx[]) const  {
                  uint64 offset = 0; 
                  for(uint i = 0; i < mNumberOfDimensions; ++i ) {
		    offset += idx[i] * mOffsetBase[i];
		  }
                  return offset;
}

template <class DataType>
void MultidimensionalArray<DataType>::precomputeVolume() {
  if(!hasValidShape()) 
          mVolume = 0;
  else {
          mVolume = 1;
          for(vector<int>::const_iterator i = mShape.begin(); i != mShape.end() ; ++i) mVolume *= *i;
  }
}


template <class DataType>
MultidimensionalArray<DataType> * MultidimensionalArray<DataType>::createNewByType(vector<int> shape, string file, const int CUBETYPE) const {
      switch (CUBETYPE) {
              case CUBETYPE_SAME://0
                return createNew(shape, file);// by default, keep the type the same
              case CUBETYPE_BTREECUBE:
                return new BTreeCube<DataType>(shape,file);
              case CUBETYPE_BIGARRAY:
                return new BigArray<DataType>(shape,file);
              case CUBETYPE_CDBCUBE:
                return new CDBCube<DataType>(shape,file);
              default:
                cerr << "BUG! Don't know cube type, will default to same"<< endl;
                return createNew(shape, file);
      }
}

template<class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr 
MultidimensionalArray<DataType>::partial_rollup(const InvertedAgMap map[], const string fname, const int CUBETYPE) const throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument) {
  // the "ag_maps" would be called "inverted ag_maps" in the litOLAP java
  // Such an inverted map says, for each base-cube attribute value, which
  // rolled-up attribute values (for the new cube) it contributes to

  // calculate new shape, based on the maximum entry in each inverted ag-map.

  // this can be substantially simplified with non-overlapping hierarchies [no loop]
  // for now, we simply detect that we COULD have done so and "warn"
  if(CUBETYPE==CUBETYPE_CDBCUBE) throw std::invalid_argument(" this method doesn't yet support the CDBCube!");
  assert(CUBETYPE!=CUBETYPE_CDBCUBE);
  bool could_have_used_simple = true;

  if(verbosepartialrollup) cout << "in C++, partial_rollup" << endl;  cout.flush();
  
  vector<int> new_shape;
  if(hasValidShape()) {
          for (uint i=0; i < mShape.size(); ++i)
            if (map[i].isIdentity()) {
              if(verbosepartialrollup) cout << " dimension " << i << " uses identity map" << endl;
              new_shape.push_back(mShape[i]);
            }
            else {
              int newdim_max = -1;
              for (uint j=0; j < (uint) mShape[i]; ++j) {
                vector<int> ancestors = map[i].getAttributeAncestors(j);
                if (ancestors.size() > 1) could_have_used_simple = false;
                for (uint k = 0; k < ancestors.size() ; ++k)
                  if ( ancestors[k] > newdim_max) newdim_max = ancestors[k];
              }
              assert(newdim_max >= 0);
              new_shape.push_back( newdim_max + 1);
              if(verbosepartialrollup) cout << " dimension " << i << " is going to have size " << newdim_max+1 << endl;
            }

          if (could_have_used_simple) 
            if(verbosepartialrollup) cout << "[Notice]: this case could make use of simplified partial_rollup" << endl;

          if(verbosepartialrollup)  {
            cout << "dimensions of rollup will be ";
            for (uint j = 0; j < new_shape.size(); ++j)  cout << new_shape[j] << " ";
            cout << endl;
          }
  } else {
          new_shape = mShape;
  }
  MultidimensionalArray<DataType> * rollup = createNewByType( new_shape, fname, CUBETYPE);
  rollup->fillWithZeroes();  

  if(verbosepartialrollup) cout << " preparing to iterate over src cube" << endl;
  std::auto_ptr<MultidimensionalIterator<DataType> > it(iterator()); 

  int new_where[mShape.size()];

   while (it->next()) {
     vector<int> where = it->key();

     
     DataType what = it->value();
 
     // initialize to max count.
     uint variable_base_ctr_max = 1;
     // if the number of possibilities for the dim_i value is p_i, then we
     // can view the "first digit" as representing dimension 0, so it has
     // values 0..p_0 -1.  The second digit represents dimension 1, so it
     // has values 0..p_1 -1 and so forth.
     uint digit_modulus[mShape.size()];

     for (uint j=0; j < mShape.size(); ++j) {
       digit_modulus[j] =  (map[j].isIdentity()) ? 1 : map[j].getNumberOfAncestors(where[j]);
       variable_base_ctr_max *= digit_modulus[j];
     }


     if(verbosepartialrollup > 5)  {
        cout << "digit moduli are " ;
        for (uint i = 0; i < mShape.size(); ++i) cout << " " << digit_modulus[i];
        cout<<endl;
        cout.flush(); 
     }
     for(uint variable_base_ctr = 0; variable_base_ctr < variable_base_ctr_max; ++variable_base_ctr) {
       double probability_product = 1.0; // implement weighting
       uint work = variable_base_ctr;  // working copy will be trashed
       for (uint digit = 0; digit < mShape.size(); ++digit) {
	 if (map[digit].isIdentity()) new_where[digit] = where[digit]; // special case of id map
	 else {
	   uint cur_digit = work % digit_modulus[digit];
	   new_where[digit] = map[digit].getAttributeAncestor(where[digit],cur_digit);
	   probability_product *= map[digit].getAttributeAncestorWeight(where[digit],cur_digit);
	   work /= digit_modulus[digit];
	 }
       }

       if(verbosepartialrollup > 5)  {
         cout << " aggregate at ";
         for (uint i = 0; i < mShape.size(); ++i)  cout << " " << new_where[i];
         cout<<endl;
         cout.flush();
       }
       // DL will either love or hate the "maybe_round" idea.  He will hate the coding, regardless!
       rollup->put( new_where,  maybe_round(probability_product*what + rollup->get( new_where),DataType(0)));
     }
   }
   //delete it;
   return rollup;
}


template<class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr  
  MultidimensionalArray<DataType>::full_rollup( const vector<int>& dimsToKill, const string fname,const int CUBETYPE) const  throw(std::out_of_range, std::ifstream::failure, std::bad_alloc,std::invalid_argument) {
  if(CUBETYPE==CUBETYPE_CDBCUBE) throw std::invalid_argument(" this method doesn't yet support the CDBCube!");
  assert(CUBETYPE!=CUBETYPE_CDBCUBE);
  vector<bool> keep(mNumberOfDimensions, true);
  for ( uint i = 0;  i < dimsToKill.size(); ++i) {
    keep.at(dimsToKill[i]) = false;
  }

  vector<int> new_shape;

   for (uint i=0; i <mNumberOfDimensions; ++i)
    if (keep[i]) new_shape.push_back(mShape[i]);

   int newdim = new_shape.size();

#if 0
  cout << " new shape is ";
  for (uint i=0; i < (uint) newdim; ++i)
    cout << " " << new_shape[i];
  cout << endl;
#endif

  // newdim is now the number of remaining dimensions

  MultidimensionalArray<DataType> * rollup = createNewByType( new_shape, fname, CUBETYPE) ;
  rollup->fillWithZeroes();  // hmmm, or is it implicit that "create new" should do this?// yes, it should be implicit, but there is little harm in leaving it there -DL
  std::auto_ptr<MultidimensionalIterator<DataType> >  it(iterator());    

   int new_where[newdim];

    while (it->next()) {
      vector<int> where = it->key();
      DataType what = it->value();

      // suspect this is faster than using an erase()
      // Daniel: come to think of it, you might be right
      int j=0;  
      for (uint i=0; i <mNumberOfDimensions; ++i) 
	if (keep[i])  new_where[j++] = (int) where[i];

      // aggregate
      rollup->put( new_where,  what + rollup->get( new_where));
    }
    //delete it;
    return rollup;
}

template<class DataType>
DataType MultidimensionalArray<DataType>::findMaxVal() const throw(std::out_of_range) {
        std::auto_ptr<MultidimensionalIterator<DataType> > it(iterator());
        DataType maxval(0);
        if(it->next())
                maxval = it->value();
        else 
                throw  std::out_of_range("[findMaxVal] your cube is empty!!!");
        while (it->next()) {
                if(it->value() > maxval) maxval = it->value();
        }
	//delete it;
        return maxval;
}
template<class DataType>
int MultidimensionalArray<DataType>::getNumberOfRecords()  {
        int counter = 0;
        std::auto_ptr<MultidimensionalIterator<DataType> > it(iterator());
        while (it->next()) {
                ++counter;
        }
	//delete it;
        return counter;
}

/**
* this may look like a silly method, but derived classes such as
* cdbcube, that have to enter a different phase when loading need
* this.
*/
template<class DataType>
void MultidimensionalArray<DataType>::load( MultidimensionalIterator<DataType>& it) {
      while (it.next())  put(it.key(),it.value());
}

//Owen: this is implemented as a straightforward copy.
//Daniel think it is a bad idea, very costly
template<class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
     MultidimensionalArray<DataType>::slice( const uint dim, const uint slice_at, const string fname ,const int CUBETYPE) const throw(std::out_of_range,std::ifstream::failure, std::bad_alloc){
   if(dim >= mNumberOfDimensions)  {
            stringstream strs;
            strs << "you only have "<<mNumberOfDimensions<<" dimensions, you can't kill dimension " << dim;
            //string errormsg;
            //strs >>  errormsg ;
            cerr << strs.str()<< endl;
            throw std::out_of_range(strs.str());
   }
   if(hasValidShape() && mShape[dim] <= (int) slice_at)  {
            stringstream strs;
            strs << "you only have "<<mShape[dim]<<" attribute values at dimension "<< dim<< ", you can't kill value " << slice_at;
            //string errormsg;
            //strs >>  errormsg ; 
            cerr << strs.str()<< endl;
            throw std::out_of_range(strs.str());
   }
   //class suriterator;
  assert(   (mShape[dim] > (int) slice_at || !hasValidShape()));
  
  vector<int> newShape(mShape);
  newShape.erase(newShape.begin()+dim);

  MultidimensionalArray<DataType> * result = createNewByType( newShape, fname, CUBETYPE);

  std::auto_ptr<MultidimensionalIterator<DataType> >  it (sliceIterator(dim,slice_at));    
  DeleteDimensionIterator<DataType> itdd(*it,dim);
  /*
  while (it->next()) {
      vector<int> where = it->key();
      DataType what = it->value();
      
      vector<int> new_where(where);
      new_where.erase(new_where.begin()+dim);  // or is Do It Yourself much cheaper?

      result->put( new_where, what);
    }*/
    result->load(itdd);
    return result;
}


/* materialize an iceberg cube.  Mapping with a unary operation that "rounds small down to 0"
   would not work once we support  "0 != undef" */

template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
       MultidimensionalArray<DataType>::iceberg( const DataType threshold,
						 const string fname,const int CUBETYPE) const throw(std::ifstream::failure, std::bad_alloc){
    
  MultidimensionalArray<DataType> * result = createNewByType( mShape, fname, CUBETYPE);
  std::auto_ptr<MultidimensionalIterator<DataType> >  base_it(iterator());
  IcebergMultidimensionalIterator<DataType> it( *base_it,threshold );

  result->load(it);
  return result;
}  
  


/* collapse this cube and a bunch of others into one new cube */
/**
* Daniel thinks this is a long and confusing function call. I will create my
* own. I also think  you shouldn't mix and match "appending and glueing".
* Anyhow, I feel I need "glue" and that's simpler than what you are trying to do
* so I will just implement it. I'm not even sure what you coded does the glue
* thing.
*/
template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
       MultidimensionalArray<DataType>::collapse( CubeSet<DataType> others,const vector<int>&  attValueStarting ,
						  const int dim, const string fname, 
						  const int CUBETYPE) const 
     throw(std::ifstream::failure, std::bad_alloc,std::invalid_argument) {
  if(CUBETYPE==CUBETYPE_CDBCUBE) throw std::invalid_argument(" this method doesn't yet support the CDBCube!");
  assert(CUBETYPE!=CUBETYPE_CDBCUBE);
   int numothers = others.size(); 
  cout << " collapse begins, numothers=" << numothers << ", dim=" << dim <<endl; 
  vector<int>resultShape(mShape);  // except on gluing dim, it's bigger
  resultShape[dim] = attValueStarting[numothers+1];  // final prefix sum = overall size

  MultidimensionalArray<DataType> * result = createNewByType( resultShape, fname, CUBETYPE);

  IteratorPtr it = iterator();    

  // all my values first.

  while (it->next()) {
      result->put( it->key(), it->value());  // no adjustment
  }
  delete it;

  for (uint i=0; i < (uint) numothers; ++i) {

    it = others.get(i).iterator();
    int offset = attValueStarting[i+1];

    // copy this guy's data
    while (it->next()) {
      vector<int> where = it->key();
      where[dim] += offset;  // adjust it
      result->put( where, it->value());  
    }
    delete it;
  }
  return result;
}


/**
* "glue" is the reverse of slice. Given a cube, we simply add new cubes to it on
* a new dimension. All cubes must have the *same* shape.
*/
template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
       MultidimensionalArray<DataType>::glue( CubeSet<DataType> others , const string fname, const int CUBETYPE) const 
     throw(std::ifstream::failure, std::bad_alloc,std::invalid_argument) {
  int numothers = others.size();
  for(int cubeindex = 0 ; cubeindex < numothers ; ++ cubeindex) 
          if(others.get(cubeindex).mShape != mShape) throw std::invalid_argument("When gluing, all cubes must have the same shape!");
  vector<int>resultShape(mShape);
  resultShape.push_back(numothers + 1);
  MultidimensionalArray<DataType> * result = createNewByType( resultShape, fname, CUBETYPE);
  IteratorSet<DataType> it;
  it.add(*iterator());
  for(int cubeindex = 0 ; cubeindex < numothers ; ++ cubeindex)
          it.add(*others.get(cubeindex).iterator());
  GlueIterator<DataType> gi(it);
  result->load(gi);
  for(int cubeindex = 0 ; cubeindex < numothers + 1 ; ++ cubeindex)
          delete &it.get(cubeindex);
  return result;
  
}


/* turn a d-dim array into d+1 by prepending a phony singleton dimension */
template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
       MultidimensionalArray<DataType>::prepend_zero_coordinate( const string fname, const int CUBETYPE) const
     throw(std::ifstream::failure, std::bad_alloc){
  vector<int>resultShape(mShape.size()+1);  
  resultShape[0]=1;  // singleton 
  for (uint i=0; i < mShape.size(); ++i)  resultShape[i+1] = mShape[i];

  MultidimensionalArray<DataType> * result = createNewByType( resultShape, fname, CUBETYPE);
  std::auto_ptr<MultidimensionalIterator<DataType> >  it(iterator());
  PrependDimensionIterator<DataType> pdi(*it,0);
  result->load(pdi);
  /*const uint resultDims = resultShape.size();
  vector<int> new_where(resultDims,0);

  while (it->next()) {
    // prepend the 0
      vector<int> where = it->key();
      for (uint i=0; i < resultDims-1; ++i)
	new_where[i+1] = where[i];
      result->put( new_where, it->value());  
  }*/
  return result;
}

template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
     MultidimensionalArray<DataType>::map_pointwise( const binary_olap_operator<DataType> &op, 
						     const MultidimensionalArray & operand2,
						     const string fname,const int CUBETYPE) const throw(std::ifstream::failure, std::bad_alloc,std::invalid_argument) {
  if(CUBETYPE==CUBETYPE_CDBCUBE) throw std::invalid_argument(" this method doesn't yet support the CDBCube!");
  assert(CUBETYPE!=CUBETYPE_CDBCUBE);
  assert( operand2.mShape == mShape);
  MultidimensionalArray<DataType> * result = createNewByType( mShape, fname, CUBETYPE);

  // due to polymorphism, it is unrealistic to assume that two iterators will
  // travel two (possibly different derived classes) in tandem.  We could optimize for
  // that case, to avoid making zillions of "random access" calls as is done here
  std::auto_ptr<MultidimensionalIterator<DataType> >  it(iterator());
   

  // pass1: scan current cube looking for its entries

  while (it->next()) {
      vector<int> where = it->key();
      DataType what = it->value();
      
      
      DataType pt_result = op(true,what, ! operand2.undefinedAt(where), operand2.get(where));
      
      
      result->put( where, pt_result);
    }
    //delete it;

    // pass2, scan operand2 looking for entries not in current cube

    it = std::auto_ptr<MultidimensionalIterator<DataType> >(operand2.iterator());    // I'm hoping the auto_ptr will do the right thing here
    
    while (it->next()) {
      vector<int> where = it->key();
      DataType what = it->value();
      
      if ( undefinedAt(where)) {
	DataType pt_result = op(false,what,true,what);
	result->put( where, pt_result);
      }
    }
    return result;
}

template <class DataType>
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr
     MultidimensionalArray<DataType>::map_1dvectors( const vector_olap_operator<DataType> &op, 
						     const MultidimensionalArray & op2,
						     const int vector_length,
						     const int component_dim,
						     const int component_dim2,
						     const int p0, int  const p1, // thresholding params
						     const string fname,const int CUBETYPE) const throw(std::ifstream::failure, std::bad_alloc,std::invalid_argument) {
  if(CUBETYPE==CUBETYPE_CDBCUBE) throw std::invalid_argument(" this method doesn't yet support the CDBCube!");
  assert(CUBETYPE!=CUBETYPE_CDBCUBE);
  assert( component_dim >= 0 && mShape.size() > (uint) component_dim);
  assert( component_dim2 >= 0 && op2.mShape.size() > (uint) component_dim2);

  assert( op2.mShape[component_dim2] == mShape[component_dim]);  // consistency in vector lengths
  assert( mShape[component_dim] == INVALID_SIZE || mShape[component_dim] == vector_length);

  // we remove the component_dim from each dimension and then collect up the
  // cross product of the remaining dimensions

  cout << "map_1dvectors: this is going to hurt unless slice iterators are good." << endl;

  int common_dim_shape = vector_length;

  vector<int> iterLimit(mShape); 
  // iterLimit[component_dim] = 1;  Buzz!

  // Bugfix:   Don't ignore collocations that exist
  // but have a 0 value stored at dimension component_dim = 0;
  // however, 
  // if the collocation has total=0 in either cube, then we do not
  // produce or use its vector.

  // Bugfix might be too expensive in the inner loop; 
  // alternative is to compute the rollup and use that to guide
  // the inner loop.

  vector<int> iterLimit2(op2.mShape); 
  // iterLimit2[component_dim2] = 1;  -- source of  the bug
  vector<int> s(mNumberOfDimensions,0);
  vector<int> newShape;
  for (uint i = 0; i < mShape.size(); ++i) // first copy
    if (i != (uint)component_dim) newShape.push_back( mShape[i]);
  for (uint i = 0; i < mShape.size(); ++i)  // second copy
    if (i != (uint)component_dim2) newShape.push_back( op2.mShape[i]);

  MultidimensionalArray<DataType> * result = createNewByType( newShape, fname, CUBETYPE);


  vector<int> extraparams(2);
  extraparams[0] = p0;
  extraparams[1] = p1;
  // nested iterators that skip the "component dim"
  IteratorPtr it = iterator(s,iterLimit);  
  while (it->next()) {

    vector<DataType> v1(common_dim_shape,0);
    
    vector<int> where = it->key();
#if 0
    cout << "outer where has ";
    for (uint i = 0; i < where.size(); ++i)
      cout << " " << where[i];
    cout << endl;
#endif

    int thisDimValue = where[component_dim];
    bool redundant = false;
    // collect the elements of component_dim
    for (uint i = 0; i < (uint) common_dim_shape; ++i) {
      where[component_dim] = (int) i;
      v1[i] = get(where);
      // have you been seen before?
      // If any component before thisDimValue
      // is nonzero, then you have!  
      if (v1[i] > 0 && i < (uint) thisDimValue) {redundant = true; break;}
    }

    //    if (redundant) cout << "redundant" << endl;

    if (redundant) continue;  // outer loop
    
    // inner loop
    IteratorPtr it2 = op2.iterator(s,iterLimit2);  
    while (it2->next()) {
      vector<DataType> v2(common_dim_shape,0);  
      
      vector<int> where2 = it2->key();
      
#if 0
      cout << "inner where has ";
      for (uint i = 0; i < where2.size(); ++i)
	cout << " " << where2[i];
      cout << endl;
#endif

      thisDimValue = where2[component_dim2];
      redundant = false;
      for (uint i = 0; i < (uint) common_dim_shape; ++i) {
	where2[component_dim2] = (int) i;
	//	cout << "getting component2 at index " << i << endl;
	v2[i] = op2.get(where2);
	if (v2[i] > 0 && i < (uint) thisDimValue) {redundant = true; break;}
      }

      if (redundant) continue;  // inner loop

      // cout << "lengths of v1, v2, extraparams are " << v1.size() << "," << v2.size() << "," << extraparams.size() << endl;

      DataType pt_result = op(v1, v2, extraparams);
      
      // surpress zeros..should be parameterized.

      if (pt_result != DataType(0)) {
	vector<int> where_to_store(newShape.size());
	int j = 0;
	for (uint i = 0; i < mShape.size(); ++i)
	  if (i != (uint) component_dim) where_to_store[j++] = where[i];
	for (uint i = 0; i < op2.mShape.size(); ++i)
	  if (i != (uint) component_dim2) where_to_store[j++] = where2[i];

#if 0		
	cout << "storing " << pt_result << " to ";
	for (uint i = 0; i < where_to_store.size(); ++i)
	  cout << " " << where_to_store[i];
#endif	  
	result->put( where_to_store, pt_result);}
    }
    delete it2;
  }
  delete it;
  return result;
}




// basically, order statistics on a cube
// returns the smallest threshold that will contain at least N items when used in an iceberg
// iterator. Note: the iceberg may return considerably more than N entries due to ties and it is
// the user's responsibility to deal with this
// Design choices: 1. binary search, multiple iceberg scans  (read-only).
// 2. Priority queue of top N.  (I choose this.)  -ofk

/**
* Good stuff, but I improved it so that the iceberg cube will have no more
* than N elements. I hope. -Daniel

*/
template<class DataType>
DataType     MultidimensionalArray<DataType>::topN_frequency(const uint n) const throw(std::out_of_range) {
  std::auto_ptr<MultidimensionalIterator<DataType> >  it(iterator());
  priority_queue<DataType> top_thresholds; // stored -x so that top() returns lowest threshold
  while (it->next()) {
    top_thresholds.push( -(it->value()));
    if (top_thresholds.size() > n + 1) top_thresholds.pop();
  }
  if (top_thresholds.empty()) {
    throw std::out_of_range("[topN_frequency] Warning, your cube is empty!!");
  }
  if (top_thresholds.size() == n+1) {
          DataType value = top_thresholds.top();
          top_thresholds.pop();
          while ( !top_thresholds.empty() ) {
                  if( top_thresholds.top() == value ) top_thresholds.pop();
                  else break;
          }
  }
  if (top_thresholds.empty()) {
            stringstream strs;
            strs << "[topN_frequency] No iceberg cube containing only "<< n << "elements.";
            throw std::out_of_range(strs.str());
  }
  return -(top_thresholds.top());
}


/****************************************************************
****************************************************************
****************************************************************
* The rest has to do with iterators
**************************************************************
****************************************************************
****************************************************************/

template <class DataType>
class MultidimensionalIterator {
  public: 
  MultidimensionalIterator(){}
    virtual bool next() = 0;
    virtual DataType value() const =0;

    virtual vector<int>& key()=0;
    virtual ~MultidimensionalIterator(){}
    virtual MultidimensionalIterator<DataType> * clone() const=0;
};


/*** 
 * cut and paste from IcebergMultidimensionalIterator; there is an alternative implementation that would use
 * an IcebergMultidimensionalIterator (first with threshold - epsilon, then with threshold, counting)
 * but then I would have to know a different epsilon for each DataType.
  
 * suspended work on this, because it tempted me to change the semantics
 * of Iceberg... to force a reset on mMI in the constructor..but perhaps Daniel wants the
 * fact that you can start iceberg filtering midway through an iteration.  -ofk
 *
 
template <class DataType>
class topNMultidimensionalIterator: public MultidimensionalIterator<DataType> {
  public: 
    topNMultidimensionalIterator(const MultidimensionalIterator<DataType>& mi, DataType threshold, const uint N): mMI(mi.clone()), mThreshold(threshold),mFirstPass(false),mN(N),mNumYielded(0){}
    
    topNMultidimensionalIterator(const topNMultidimensionalIterator<DataType>& other):mMI(other.mMI->clone()),mThreshold(other.mThreshold),mN(other.mN), mFirstPass(other.mFirstPass),mNumYielded(other.mNumYielded) {}
    inline bool next() {
            if(verbose) cout <<"calling topN next..."<<endl;
            while(mMI->next()) {
              if(verbose) cout <<"mMI.value() = "<<mMI->value()<<endl;
              if(mMI->value() >= mThreshold)  {
                      if(verbose) cout <<"iceberg next...returning true"<<endl;
                      return true;
              }
             }
             if(verbose) cout <<"iceberg next...returning false"<<endl;
             return false; };
    inline DataType value() const {return mMI->value();}
    inline vector<int>& key() {
    if(careful) {
      vector<int> key = mMI->key();
      for(uint k = 0; k < key.size(); ++k) {
        if(key[k] <0) cerr << "trouble in C++ land with iceberg iterator." << endl;
      }
    }
    return mMI->key();
    };
    virtual ~IcebergMultidimensionalIterator(){delete mMI;}
    
    IcebergMultidimensionalIterator<DataType> * clone() const {return new IcebergMultidimensionalIterator<DataType>(*this);}
  private: 
  enum{verbose = false,careful=false};
    MultidimensionalIterator<DataType> * mMI;
    const DataType mThreshold;
};
--- end of suspended-work area
*/


template <class DataType>
class IcebergMultidimensionalIterator: public MultidimensionalIterator<DataType> {
  public: 
    IcebergMultidimensionalIterator(const MultidimensionalIterator<DataType>& mi, DataType threshold): mMI(mi.clone()), mThreshold(threshold){}
    
    IcebergMultidimensionalIterator(const IcebergMultidimensionalIterator<DataType>& other):mMI(other.mMI->clone()),mThreshold(other.mThreshold) {}
    inline bool next() {
            if(verbose) cout <<"calling iceberg next..."<<endl;
            while(mMI->next()) {
              if(verbose) cout <<"mMI.value() = "<<mMI->value()<<endl;
              if(mMI->value() >= mThreshold)  {
                      if(verbose) cout <<"iceberg next...returning true"<<endl;
                      return true;
              }
             }
             if(verbose) cout <<"iceberg next...returning false"<<endl;
             return false; };
    inline DataType value() const {return mMI->value();}
    inline vector<int>& key() {
    return mMI->key();
    };
    virtual ~IcebergMultidimensionalIterator(){delete mMI;}
    
    IcebergMultidimensionalIterator<DataType> * clone() const {return new IcebergMultidimensionalIterator<DataType>(*this);}
  private: 
  enum{verbose = false};
    MultidimensionalIterator<DataType> * mMI;
    const DataType mThreshold;
};


template <class DataType>
class BufferedIterator {
  public: 
    BufferedIterator(const MultidimensionalIterator<DataType>& mi): mMI(mi.clone()), needToCallNext(1), cachedNextResult(0){}
    
    BufferedIterator(const BufferedIterator<DataType>& other):mMI(other.mMI->clone()), needToCallNext(other.needToCallNext), cachedNextResult(other.cachedNextResult) {}
    inline bool hasNext() {
	    if (needToCallNext) {
		cachedNextResult = mMI->next();
		needToCallNext = false;
	    }
	    return cachedNextResult;        
    }
    inline void next() throw(std::out_of_range) {
            if ( ! hasNext()) throw std::out_of_range("You called next() but hasNext() is false!");
            needToCallNext = true;
    }
    
    
    inline DataType value() const {return mMI->value();}

    inline vector<int>& key(){return mMI->key();}
    virtual ~BufferedIterator(){}

  private: 

    MultidimensionalIterator<DataType> * mMI;
    bool needToCallNext, cachedNextResult;
};

/**
* This wrapper simply deletes attribute values of one dimension. Useful
* for slicing.
*/
template <class DataType>
class DeleteDimensionIterator: public MultidimensionalIterator<DataType> {
  public: 
    DeleteDimensionIterator(const MultidimensionalIterator<DataType>& mi, int dim): mMI(mi.clone()), mDim(dim),mNewWhere(0){}
    
    DeleteDimensionIterator(const DeleteDimensionIterator<DataType>& other):mMI(other.mMI->clone()),mDim(other.mDim),mNewWhere(other.mNewWhere) {}
    inline bool next() {
            return mMI->next();
    };
    inline DataType value() const {return mMI->value();}
    inline vector<int>& key() {
       // next line should  be elsewhere, really
       if(mNewWhere.size() == 0) mNewWhere.resize(mMI->key().size() - 1);
       // we can't do it much faster than this:
       memcpy(&mNewWhere[0],&mMI->key()[0],mDim*sizeof(int));
       memcpy(&mNewWhere[mDim],&mMI->key()[mDim+1],(mNewWhere.size()-mDim)*sizeof(int));
      return mNewWhere; 
      //return mNewWhere;
    };
    virtual ~DeleteDimensionIterator(){delete mMI;}
    
    DeleteDimensionIterator<DataType> * clone() const {return new DeleteDimensionIterator<DataType>(*this);}
  private: 
    MultidimensionalIterator<DataType> * mMI;
    const int mDim;
    vector<int> mNewWhere;
};


/**
* This wrapper "glues" several iterators.
*/
template <class DataType>
class GlueIterator: public MultidimensionalIterator<DataType> {
  public: 
    GlueIterator(IteratorSet<DataType> mi): mMI(mi), mNumber(mMI.size()), mNewWhere(0), mIteratorCounter(0){
            
    }
    
    GlueIterator(const GlueIterator<DataType>& other):mMI(other.mMI),mNumber(other.mNumber),mNewWhere(other.mNewWhere), mIteratorCounter(other.mIteratorCounter) {}
    inline bool next() {
            if(mIteratorCounter + 1> mNumber) return false;
            if( mMI.get(mIteratorCounter).next()) return true;
            ++mIteratorCounter;
            return next();
    }
    inline DataType value() const {
          assert(mIteratorCounter < mNumber);
          return mMI.get(mIteratorCounter).value();
    }
    inline vector<int>& key() {
            assert(mIteratorCounter < mNumber);
            // this could be optimized, use memcpy
            mNewWhere = mMI.get(mIteratorCounter).key();
            mNewWhere.push_back(mIteratorCounter);
            return mNewWhere;
    }
    virtual ~GlueIterator(){}
    
    GlueIterator<DataType> * clone() const {return new GlueIterator<DataType>(*this);}
  private: 
    IteratorSet<DataType> mMI;
    int mNumber;
    vector<int> mNewWhere;
    int mIteratorCounter;
};

/**
* This wrapper simply add a dimension 0 with the specified attr. value.
*/
template <class DataType>
class PrependDimensionIterator: public MultidimensionalIterator<DataType> {
  public: 
    PrependDimensionIterator(const MultidimensionalIterator<DataType>& mi, int AttributeValue): mMI(mi.clone()), mAttributeValue(AttributeValue), mNewWhere(0){}
    
    PrependDimensionIterator(const PrependDimensionIterator<DataType>& other):mMI(other.mMI->clone()),mAttributeValue(other.mAttributeValue),mNewWhere(other.mNewWhere) {}
    inline bool next() {
            return mMI->next();
    };
    inline DataType value() const {return mMI->value();}
    inline vector<int>& key() {
            // this could be optimized, use memcpy
            mNewWhere = mMI->key();
            mNewWhere.insert(mNewWhere.begin(),mAttributeValue);
            return mNewWhere;
    };
    virtual ~PrependDimensionIterator(){delete mMI;}
    
    PrependDimensionIterator<DataType> * clone() const {return new PrependDimensionIterator<DataType>(*this);}
  private: 
    MultidimensionalIterator<DataType> * mMI;
    const int mAttributeValue;
    vector<int> mNewWhere;
};


/**
* The goal of this class is to provide a simple data structure to hold cubes
* while avoiding visible pointers.  However, the user must ensure that the
* items in the set continue to exist during the time the items are in the
* set, or we will have "dangling pointers".  OFK wonders if this could be 
* avoided by a "virtual copy constructor"  like "clone()" required for all
* classes deriving from MultidimensionalArray.
*/
template <class DataType>
class CubeSet {
  public:
  CubeSet(): mCubes(){}
  ~CubeSet(){}
  CubeSet(const CubeSet<DataType>& cs): mCubes(cs.mCubes){}
  void add(MultidimensionalArray<DataType> & ma) {mCubes.push_back(&ma);}
  void remove(int i) {
    mCubes.erase(mCubes.begin()+i);
  }
  int size() const {return (int) mCubes.size();}
  MultidimensionalArray<DataType> & get(int i) const {return *mCubes[i];}
  private:
    vector<MultidimensionalArray<DataType> *> mCubes;// we hide this ugliness
};


/**
* The goal of this class is to provide a simple data structure to hold iterators
* while avoiding pointers.
*/
template <class DataType>
class IteratorSet {
  public:
  IteratorSet(): mIterator(){}
  ~IteratorSet(){}
  IteratorSet(const IteratorSet<DataType>& cs): mIterator(cs.mIterator){}
  void add(MultidimensionalIterator<DataType> & ma) {mIterator.push_back(&ma);}
  void remove(int i) {
    mIterator.erase(mIterator.begin()+i);
  }
  
  int size() const {return (int) mIterator.size();}
  MultidimensionalIterator<DataType> & get(int i) const {return *mIterator[i];}
  private:
    vector<MultidimensionalIterator<DataType> *> mIterator;// this is too ugly
};

#endif 
