 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef OLAP_OPERATOR_H
#define OLAP_OPERATOR_H


#include <vector>
#include <stdexcept>
#include <memory>
#include <string>

using namespace std;


template<class DataType>
class Sum;

template<class DataType>
class Diff;

template<class DataType>
class ValueOnlyInFirst;

template<class DataType>
class ValueExistsInSecond;

template<class DataType>
class ScaledIcebergedCosineSqr;


/**
* this is the base class for binary olap operators
*/
template<class DataType>
struct binary_olap_operator{
  virtual DataType operator() (const bool v1_exists, const DataType &v1, 
			       const bool v2_exists, const DataType &v2) const = 0; 
};

/**
* This is the factory: no need to mix it with the base class.
*/
template<class DataType>
struct binary_olap_operator_factory{
  // static members are faster and they avoid useless object creation
  static Sum<DataType> mSum;
  static Diff<DataType> mDiff;
  static ValueOnlyInFirst<DataType> mValueOnlyInFirst;
  static ValueExistsInSecond<DataType> mValueExistsInSecond;
  
  static binary_olap_operator<DataType> & makeOperator( const char sym) {
    switch (sym) {
    case '+' : return mSum;
    case '-' : return mDiff;
    case '\\' : return mValueOnlyInFirst;
    case '/' : return mValueExistsInSecond;
    default: cerr << "Unknown binary_olap_operator, defaulting to Sum: " << sym << endl;
      return mSum;
    }  
  }
};
/**
* daniel: we now build the static members used by the factory.
*/
template<class DataType>
Sum<DataType> binary_olap_operator_factory<DataType>::mSum;
template<class DataType>
Diff<DataType> binary_olap_operator_factory<DataType>::mDiff;
template<class DataType>
ValueOnlyInFirst<DataType> binary_olap_operator_factory<DataType>::mValueOnlyInFirst;
template<class DataType>
ValueExistsInSecond<DataType> binary_olap_operator_factory<DataType>::mValueExistsInSecond;
  
template<class DataType> struct Sum : public binary_olap_operator<DataType>{
  DataType operator() (const bool v1_exists, const DataType &v1, 
		       const bool v2_exists, const DataType &v2) const { 
    return (v1_exists ? v1 : DataType(0)) + (v2_exists ? v2 : DataType(0));}
};

template<class DataType> struct Diff : public binary_olap_operator<DataType>{
  DataType operator() (const bool v1_exists, const DataType &v1, 
		       const bool v2_exists, const DataType &v2) const { 
    return (v1_exists ? v1 : DataType(0)) - (v2_exists ? v2 : DataType(0));}
};

// cell 1 = in first, not second;  cell 0 = in second.  Probably caller will make "in none" = undefined
template<class DataType> struct ValueOnlyInFirst: public binary_olap_operator<DataType>{
  DataType operator() (const bool v1_exists, const DataType &v1, 
		       const bool v2_exists, const DataType &v2) const { 
    if (v1_exists && ! v2_exists) return DataType(1);
    return DataType(0);
  }
};

// preserves measure of first cube. Form of "intersect"
template<class DataType> struct ValueExistsInSecond: public binary_olap_operator<DataType>{
  DataType operator() (const bool v1_exists, const DataType &v1, 
		       const bool v2_exists, const DataType &v2) const { 
    if (v1_exists && v2_exists) return v1;
    return DataType(0);
  }
};

/**
* this is the base class for binary vector olap operators
*/
template<class DataType>
struct vector_olap_operator{
  virtual DataType operator() (const vector<DataType> &v1, 
			       const vector<DataType> &v2, const vector<int> &otherparams) const = 0; 
};

/**
* This is the factory
*/
template<class DataType>
struct vector_olap_operator_factory{
  static ScaledIcebergedCosineSqr<DataType> mSICS;
  static vector_olap_operator<DataType> & makeOperator( const char sym) {
    switch (sym) {
    case 'c' : return mSICS;
    default: cerr << "Unknown vector_olap_operator, defaulting to SICS: " << sym << endl;
      return mSICS;
    }  
  }
};

template<class DataType>
ScaledIcebergedCosineSqr<DataType> vector_olap_operator_factory<DataType>::mSICS;

// note: we have not taken logs prior to building the vectors, unlike Turney
// we might want to adjust this accordingly.

  
template<class DataType> struct ScaledIcebergedCosineSqr : public vector_olap_operator<DataType>{
  DataType operator()(const vector<DataType> &v1, const vector<DataType> &v2, 
		      const vector<int> &params ) const {

    // Double (not DataType) due to overflows when instantiated to int
    double sum1(0);
    double sum2(0);
    double sum12(0);

    DataType sumV1(0);
    DataType sumV2(0);

    for(uint i = 0; i < v1.size(); ++i) {
      sum1 += v1[i] * v1[i];
      sum2 += v2[i] * v2[i];
      sum12 += v1[i] * v2[i];
      sumV1 += v1[i];
      sumV2 += v2[i];
    }

#if 0
    cout << "invoke SICS v1 = ";
    for (uint i=0; i < v1.size(); ++i)
      cout << " " << v1[i];
    cout << endl;
    cout << " v2 = ";
    for (uint i=0; i < v2.size(); ++i)
      cout << " " << v2[i];
    cout << endl;
    cout << "sum1=" << sum1 <<", sum2=" << sum2 << ",sum12=" << sum12 << 
       ", sumV1=" << sumV1 << ", sumV2=" << sumV2 << endl;
#endif

    // a hack....
    // pre-threshold: if either vector does not have enough weight
    // then do not claim a result.
    if (sumV1 <= DataType(params[0]) || sumV2 <= DataType(params[0]))
      return DataType(0);

    // template may instantiate dangerously for doubles and floats.

    // unsure of reasonable epsilon below
    if(sum1 * sum2 < 0.0000001 ) return DataType(0);
    DataType scaledAns;
    // for now, scaling by 1000; thresholding by params[1] (900 is good)
    // based on the observation that long vectors with fairly
    // different "angles" appear to have similar cosines.

    scaledAns = (DataType) (1000 *  sum12 * sum12 / (sum1*sum2)); 
    //    cout << "scaled answer is " << scaledAns << endl;

    if (scaledAns <= DataType(params[1])) return DataType(0); // iceberg
    return scaledAns;
  }
};

#endif
